import logging
from typing import Callable, Iterable, Union

from ..engine import Engine, SingleCoreEngine
from ..history import HistoryOptions
from ..problem import Problem
from ..result import Result
from ..startpoint import StartpointMethod, to_startpoint_method, uniform
from ..store import autosave
from .optimizer import Optimizer, ScipyOptimizer
from .options import OptimizeOptions
from .task import OptimizerTask
from .util import (
    assign_ids,
    bound_n_starts_from_env,
    postprocess_hdf5_history,
    preprocess_hdf5_history,
)

logger = logging.getLogger(__name__)


def minimize(
    problem: Problem,
    optimizer: Optimizer = None,
    n_starts: int = 100,
    ids: Iterable[str] = None,
    startpoint_method: Union[StartpointMethod, Callable, bool] = None,
    result: Result = None,
    engine: Engine = None,
    progress_bar: bool = True,
    options: OptimizeOptions = None,
    history_options: HistoryOptions = None,
    filename: Union[str, Callable, None] = None,
    overwrite: bool = False,
) -> Result:
    """
    Do multistart optimization.

    Parameters
    ----------
    problem:
        The problem to be solved.
    optimizer:
        The optimizer to be used n_starts times.
    n_starts:
        Number of starts of the optimizer.
    ids:
        Ids assigned to the startpoints.
    startpoint_method:
        Method for how to choose start points. False means the optimizer does
        not require start points, e.g. for the 'PyswarmOptimizer'.
    result:
        A result object to append the optimization results to. For example,
        one might append more runs to a previous optimization. If None,
        a new object is created.
    engine:
        Parallelization engine. Defaults to sequential execution on a
        SingleCoreEngine.
    progress_bar:
        Whether to display a progress bar.
    options:
        Various options applied to the multistart optimization.
    history_options:
        Optimizer history options.
    filename:
        Name of the hdf5 file, where the result will be saved. Default is
        None, which deactivates automatic saving. If set to
        "Auto" it will automatically generate a file named
        `year_month_day_profiling_result.hdf5`.
        Optionally a method, see docs for `pypesto.store.auto.autosave`.
    overwrite:
        Whether to overwrite `result/optimization` in the autosave file
        if it already exists.

    Returns
    -------
    result:
        Result object containing the results of all multistarts in
        `result.optimize_result`.
    """
    # optimizer
    if optimizer is None:
        optimizer = ScipyOptimizer()

    # number of starts
    n_starts = bound_n_starts_from_env(n_starts)

    # startpoint method
    if startpoint_method is None:
        startpoint_method = uniform
    # convert startpoint method to class instance
    startpoint_method = to_startpoint_method(startpoint_method)

    # check options
    if options is None:
        options = OptimizeOptions()
    options = OptimizeOptions.assert_instance(options)

    # history options
    if history_options is None:
        history_options = HistoryOptions()
    history_options = HistoryOptions.assert_instance(history_options)

    # assign startpoints
    startpoints = startpoint_method(
        n_starts=n_starts,
        problem=problem,
    )

    ids = assign_ids(
        n_starts=n_starts,
        ids=ids,
        result=result,
    )

    # prepare result
    if result is None:
        result = Result(problem)

    # engine
    if engine is None:
        engine = SingleCoreEngine()

    # change to one hdf5 storage file per start if parallel and if hdf5
    history_file = history_options.storage_file
    history_requires_postprocessing = preprocess_hdf5_history(
        history_options, engine
    )

    # define tasks
    tasks = []
    for startpoint, id in zip(startpoints, ids):
        task = OptimizerTask(
            optimizer=optimizer,
            problem=problem,
            x0=startpoint,
            id=id,
            history_options=history_options,
            optimize_options=options,
        )
        tasks.append(task)

    # perform multistart optimization
    ret = engine.execute(tasks, progress_bar=progress_bar)

    # merge hdf5 history files
    if history_requires_postprocessing:
        postprocess_hdf5_history(ret, history_file, history_options)

    # aggregate results
    for optimizer_result in ret:
        result.optimize_result.append(optimizer_result)

    # sort by best fval
    result.optimize_result.sort()

    # if history file provided, set storage file to that one
    if filename == "Auto" and history_file is not None:
        filename = history_file
    autosave(
        filename=filename,
        result=result,
        store_type="optimize",
        overwrite=overwrite,
    )

    return result



import logging
from typing import Callable

import numpy as np

from ..C import GRAD
from ..optimize import OptimizeOptions, Optimizer
from ..problem import Problem
from ..result import OptimizerResult, ProfilerResult
from .options import ProfileOptions

logger = logging.getLogger(__name__)


def walk_along_profile(
    current_profile: ProfilerResult,
    problem: Problem,
    par_direction: int,
    optimizer: Optimizer,
    options: ProfileOptions,
    create_next_guess: Callable,
    global_opt: float,
    i_par: int,
) -> ProfilerResult:
    """
    Compute half a profile.

    Walk ahead in positive direction until some stopping criterion is
    fulfilled. A two-sided profile is obtained by flipping the profile
    direction.

    Parameters
    ----------
    current_profile:
        The profile which should be computed
    problem:
        The problem to be solved.
    par_direction:
        Indicates profiling direction (+1, -1: ascending, descending)
    optimizer:
        The optimizer to be used along each profile.
    global_opt:
        log-posterior value of the global optimum
    options:
        Various options applied to the profile optimization.
    create_next_guess:
        Handle of the method which creates the next profile point proposal
    i_par:
        index for the current parameter

    Returns
    -------
    current_profile:
        The current profile, modified in-place.
    """
    # create variables which are needed during iteration
    stop_profile = False

    # while loop for profiling (will be exited by break command)
    while True:
        # get current position on the profile path
        x_now = current_profile.x_path[:, -1]

        # check if the next profile point needs to be computed
        if par_direction == -1:
            stop_profile = (x_now[i_par] <= problem.lb_full[[i_par]]) or (
                current_profile.ratio_path[-1] < options.ratio_min
            )

        if par_direction == 1:
            stop_profile = (x_now[i_par] >= problem.ub_full[[i_par]]) or (
                current_profile.ratio_path[-1] < options.ratio_min
            )

        if stop_profile:
            break

        # compute the new start point for optimization
        x_next = create_next_guess(
            x_now,
            i_par,
            par_direction,
            options,
            current_profile,
            problem,
            global_opt,
        )

        # fix current profiling parameter to current value and set
        # start point
        problem.fix_parameters(i_par, x_next[i_par])
        startpoint = np.array([x_next[i] for i in problem.x_free_indices])

        # run optimization
        # IMPORTANT: This optimization will need a proper exception
        # handling (coming soon)
        if startpoint.size > 0:
            optimizer_result = optimizer.minimize(
                problem=problem,
                x0=startpoint,
                id='0',
                optimize_options=OptimizeOptions(allow_failed_starts=False),
            )
        else:
            # if too many parameters are fixed, there is nothing to do ...
            fval = problem.objective(np.array([]))
            optimizer_result = OptimizerResult(
                id='0',
                x=np.array([]),
                fval=fval,
                n_fval=0,
                n_grad=0,
                n_res=0,
                n_hess=0,
                n_sres=0,
                x0=np.array([]),
                fval0=fval,
                time=0,
            )
            optimizer_result.update_to_full(problem=problem)

        if optimizer_result[GRAD] is not None:
            gradnorm = np.linalg.norm(
                optimizer_result[GRAD][problem.x_free_indices]
            )
        else:
            gradnorm = np.nan

        current_profile.append_profile_point(
            x=optimizer_result.x,
            fval=optimizer_result.fval,
            ratio=np.exp(global_opt - optimizer_result.fval),
            gradnorm=gradnorm,
            time=optimizer_result.time,
            exitflag=optimizer_result.exitflag,
            n_fval=optimizer_result.n_fval,
            n_grad=optimizer_result.n_grad,
            n_hess=optimizer_result.n_hess,
        )

    # free the profiling parameter again
    problem.unfix_parameters(i_par)

    return current_profile



"""DynestySampler class."""

# TODO set `nlive` to > a few * n_params**2
#      - https://github.com/joshspeagle/dynesty/issues/301#issuecomment-888461184
#      - seems easy for `NestedSampler(..., live_points=...)`, and for
#        `DynamicNestedSampler(..., nlive0=...)`.
# TODO set one of the initial live points to the best vector from optimization
#      - this should hopefully ensure a good estimate for `L_max`, used for
#        stopping criteria
#      - seems easy for `NestedSampler(..., live_points)`, but not for
#        `DynamicNestedSampler`.
# TODO allow selection of bounding methods
#      - which bounding method works well for high-dimension problems?
#        - possibly overlapping balls/cubes (reduced tuning parameters), or
#          unit cube (but only if unimodal?)
#          - ellipsoids seem to be recommended
#        - requires benchmarking...
# TODO consider "inner" sampling method
#      - Hamiltonian for large problems (D > 20) with gradients
#        - need to reshape gradients to unit cube
#        - need to provide gradients, hopefully without additional obj call
#        - or use `compute_jac` option, probably expensive
#      - multivariate slice for large problems (D > 20) without gradients
# TODO consider reflective boundaries

from __future__ import annotations

import logging
from typing import List, Union

import numpy as np

from ..problem import Problem
from ..result import McmcPtResult
from .sampler import Sampler, SamplerImportError

logger = logging.getLogger(__name__)


class DynestySampler(Sampler):
    """Use dynesty for sampling.

    Wrapper around https://dynesty.readthedocs.io/en/stable/, see there for
    details.
    """

    def __init__(
        self,
        sampler_args: dict = None,
        run_args: dict = None,
        dynamic: bool = True,
    ):
        """
        Initialize sampler.

        Parameters
        ----------
        sampler_args:
            Further keyword arguments that are passed on to the `__init__`
            method of the dynesty sampler.
        run_args:
            Further keyword arguments that are passed on to the `run_nested`
            method of the dynesty sampler.
        """
        # check dependencies
        try:
            import dynesty
        except ImportError:
            raise SamplerImportError("dynesty")

        super().__init__()

        self.dynamic = dynamic

        if sampler_args is None:
            sampler_args = {}
        self.sampler_args: dict = sampler_args

        if run_args is None:
            run_args = {}
        self.run_args: dict = run_args

        # set in initialize
        self.problem: Union[Problem, None] = None
        self.sampler: Union[dynesty.DynamicNestedSampler, None] = None

    def prior_transform(self, prior_sample: np.ndarray) -> np.ndarray:
        """Transform prior sample from unit cube to pyPESTO prior.

        TODO support priors that are not uniform.
             raise warning in `self.initialize` for now.

        Parameters
        ----------
        prior_sample:
            The prior sample, provided by dynesty.
        problem:
            The pyPESTO problem.

        Returns
        -------
        The transformed prior sample.
        """
        return (
            prior_sample * (self.problem.ub - self.problem.lb)
            + self.problem.lb
        )

    def initialize(
        self,
        problem: Problem,
        x0: Union[np.ndarray, List[np.ndarray]],
    ) -> None:
        """Initialize the sampler."""
        import dynesty

        self.problem = problem

        def loglikelihood(x):
            """Log-probability density function."""
            # check if parameter lies within bounds
            if any(x < self.problem.lb) or any(x > self.problem.ub):
                return -np.inf
            # invert sign
            # TODO this is possibly the posterior if priors are defined
            return -1.0 * self.problem.objective(x)

        sampler_class = dynesty.NestedSampler
        if self.dynamic:
            sampler_class = dynesty.DynamicNestedSampler

        # initialize sampler
        self.sampler = sampler_class(
            loglikelihood=loglikelihood,
            prior_transform=self.prior_transform,
            ndim=len(self.problem.x_free_indices),
            **self.sampler_args,
        )

        # TODO somehow use optimized vector as one of the initial live points

    def sample(self, n_samples: int, beta: float = None) -> None:
        """Return the most recent sample state."""
        if n_samples is not None:
            logger.warning(
                "`n_samples` was specified but this is incompatible with "
                "`dynesty` samplers, as they run until other stopping "
                "criteria are satisfied."
            )
        if beta is not None:
            logger.warning(
                "The temperature of a `dynesty` sampler was set, but this is "
                "irrelevant for `dynesty` samplers."
            )

        self.sampler.run_nested(**self.run_args)
        self.results = self.sampler.results

    def get_samples(self) -> McmcPtResult:
        """Get the samples into the fitting pypesto format."""
        # all walkers are concatenated, yielding a flat array
        trace_x = np.array([self.results.samples])
        trace_neglogpost = -np.array([self.results.logl])
        # the sampler uses custom adaptive priors
        trace_neglogprior = np.full(trace_neglogpost.shape, np.nan)
        # the sampler uses temperature 1
        betas = np.array([1.0])

        result = McmcPtResult(
            trace_x=trace_x,
            trace_neglogpost=trace_neglogpost,
            trace_neglogprior=trace_neglogprior,
            betas=betas,
        )

        return result



# noqa: D400,D205
"""
Optimize
========

Multistart optimization with support for various optimizers.
"""

from .ess import ESSOptimizer
from .load import (
    fill_result_from_history,
    optimization_result_from_history,
    read_result_from_file,
    read_results_from_file,
)
from .optimize import minimize
from .optimizer import (
    CmaesOptimizer,
    DlibOptimizer,
    FidesOptimizer,
    IpoptOptimizer,
    NLoptOptimizer,
    Optimizer,
    PyswarmOptimizer,
    PyswarmsOptimizer,
    ScipyDifferentialEvolutionOptimizer,
    ScipyOptimizer,
)
from .options import OptimizeOptions



"""History utility functions."""

import numbers
from typing import Dict, Sequence, Union

import numpy as np

from ..C import SUFFIXES

ResultDict = Dict[str, Union[float, np.ndarray]]
MaybeArray = Union[np.ndarray, 'np.nan']


class HistoryTypeError(ValueError):
    """Error raised when an unsupported history type is requested."""

    def __init__(self, history_type: str):
        super().__init__(
            f"Unsupported history type: {history_type}, expected {SUFFIXES}"
        )


class CsvHistoryTemplateError(ValueError):
    """Error raised when no template is given for CSV history."""

    def __init__(self, storage_file: str):
        super().__init__(
            "CSV History requires an `{id}` template in the `storage_file`, "
            f"but is {storage_file}"
        )


def trace_wrap(f):
    """
    Wrap around trace getters.

    Transform input `ix` vectors to a valid index list, and reduce for
    integer `ix` the output to a single value.
    """

    def wrapped_f(
        self, ix: Union[Sequence[int], int, None] = None, trim: bool = False
    ) -> Union[Sequence[Union[float, MaybeArray]], Union[float, MaybeArray]]:
        # whether to reduce the output
        reduce = isinstance(ix, numbers.Integral)
        # default: full list
        if ix is None:
            if trim:
                ix = self.get_trimmed_indices()
            else:
                ix = np.arange(0, len(self), dtype=int)
        # turn every input into an index list
        if reduce:
            ix = np.array([ix], dtype=int)
        # obtain the trace
        trace = f(self, ix)
        # reduce the output
        if reduce:
            trace = trace[0]
        return trace

    return wrapped_f



"""Calculate different diagnostics of the sampling result."""
import logging

import numpy as np

from ..result import Result
from .auto_correlation import autocorrelation_sokal
from .geweke_test import burn_in_by_sequential_geweke

logger = logging.getLogger(__name__)


def geweke_test(result: Result, zscore: float = 2.0) -> int:
    """
    Calculate the burn-in of MCMC chains.

    Parameters
    ----------
    result:
        The pyPESTO result object with filled sample result.
    zscore:
        The Geweke test threshold.

    Returns
    -------
    burn_in:
        Iteration where the first and the last fraction of the chain
        do not differ significantly regarding Geweke test -> Burn-In

    """
    # Get parameter samples as numpy arrays
    chain = np.asarray(result.sample_result.trace_x[0])

    # Calculate burn in index
    burn_in = burn_in_by_sequential_geweke(chain=chain, zscore=zscore)

    # Log
    logger.info(f'Geweke burn-in index: {burn_in}')

    # Fill in burn-in value into result
    result.sample_result.burn_in = burn_in

    return burn_in


def auto_correlation(result: Result) -> float:
    """
    Calculate the autocorrelation of the MCMC chains.

    Parameters
    ----------
    result:
        The pyPESTO result object with filled sample result.

    Returns
    -------
    auto_correlation:
        Estimate of the integrated autocorrelation time of
        the MCMC chains.
    """
    # Check if burn in index is available
    if result.sample_result.burn_in is None:
        geweke_test(result)

    # Get burn in index
    burn_in = result.sample_result.burn_in

    # Get chain length
    chain_length = result.sample_result.trace_x.shape[1]

    if burn_in == chain_length:
        logger.warning(
            "The autocorrelation can not "
            "be estimated. The chain seems to "
            "not have converged yet.\n"
            "You may want to use a larger number "
            "of samples."
        )
        return None

    # Get converged parameter samples as numpy arrays
    chain = np.asarray(result.sample_result.trace_x[0, burn_in:, :])

    # Calculate autocorrelation
    auto_correlation_vector = autocorrelation_sokal(chain=chain)

    # Take the maximum over all components
    _auto_correlation = max(auto_correlation_vector)

    # Log
    logger.info(f'Estimated chain autocorrelation: {_auto_correlation}')

    # Fill in autocorrelation value into result
    result.sample_result.auto_correlation = _auto_correlation

    return _auto_correlation


def effective_sample_size(result: Result) -> float:
    """
    Calculate the effective sample size of the MCMC chains.

    Parameters
    ----------
    result:
        The pyPESTO result object with filled sample result.

    Returns
    -------
    ess:
        Estimate of the effective sample size of
        the MCMC chains.
    """
    # Check if autocorrelation is available
    if result.sample_result.auto_correlation is None:
        # Calculate autocorrelation
        auto_correlation(result)

    # Get burn in index
    burn_in = result.sample_result.burn_in

    # Get estimated chain autocorrelation
    _auto_correlation = result.sample_result.auto_correlation

    if _auto_correlation is None:
        return None

    # Get converged parameter samples as numpy arrays
    chain = np.asarray(result.sample_result.trace_x[0, burn_in:, :])

    # Get length of the converged chain
    N = chain.shape[0]

    # Calculate effective sample size
    ess = N / (1 + _auto_correlation)

    # Log
    logger.info(f'Estimated effective sample size: {ess}')

    # Fill in effective sample size value into result
    result.sample_result.effective_sample_size = ess

    return ess



import logging

import numpy as np

import pypesto.optimize

from ..engine import Task
from ..history import HistoryOptions
from ..problem import Problem
from ..result import OptimizerResult

logger = logging.getLogger(__name__)


class OptimizerTask(Task):
    """A multistart optimization task, performed in `pypesto.minimize`."""

    def __init__(
        self,
        optimizer: 'pypesto.optimize.Optimizer',
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions,
        optimize_options: 'pypesto.optimize.OptimizeOptions',
    ):
        """Create the task object.

        Parameters
        ----------
        optimizer:
            The optimizer to use.
        problem:
            The problem to solve.
        x0:
            The point from which to start.
        id:
            The multistart id.
        options:
            Options object applying to optimization.
        history_options:
            Optimizer history options.
        """
        super().__init__()

        self.optimizer = optimizer
        self.problem = problem
        self.x0 = x0
        self.id = id
        self.optimize_options = optimize_options
        self.history_options = history_options

    def execute(self) -> OptimizerResult:
        """Execute the task."""
        logger.debug(f"Executing task {self.id}.")

        optimizer_result = self.optimizer.minimize(
            problem=self.problem,
            x0=self.x0,
            id=self.id,
            history_options=self.history_options,
            optimize_options=self.optimize_options,
        )
        optimizer_result.optimizer = str(self.optimizer)

        if not self.optimize_options.report_hess:
            optimizer_result.hess = None
        if not self.optimize_options.report_sres:
            optimizer_result.sres = None
        return optimizer_result



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence, Tuple

import matplotlib.pyplot as plt
import numpy as np

from .clust_color import RGBA

if TYPE_CHECKING:
    try:
        import umap

        UmapTypeObject = umap.umap_.UMAP
    except ImportError:
        UmapTypeObject = None


def projection_scatter_umap(
    umap_coordinates: np.ndarray, components: Sequence[int] = (0, 1), **kwargs
):
    """
    Plot a scatter plots for UMAP coordinates.

    Creates either one or multiple scatter plots, depending on the number of
    coordinates passed to it.

    Parameters
    ----------
    umap_coordinates:
        array of umap coordinates (returned as first output by the routine
        get_umap_representation) to be shown as scatter plot

    components:
        Components to be plotted (corresponds to columns of umap_coordinates)

    Returns
    -------
    axs:
        Either one axes object, or a dictionary of plot axes (depending on the
        number of coordinates passed)
    """
    n_components = len(components)
    if n_components == 2:
        # handle components
        x_label = f'UMAP component {components[0] + 1}'
        y_label = f'UMAP component {components[1] + 1}'
        dataset = umap_coordinates[:, components]

        # call lowlevel routine
        return ensemble_scatter_lowlevel(
            dataset, x_label=x_label, y_label=y_label, **kwargs
        )
    else:
        # We got more than two components. Plot a cross-classification table
        # Create the labels first
        component_labels = [
            f'UMAP component {components[i_comp] + 1}'
            for i_comp in range(n_components)
        ]
        # reduce pca components
        dataset = umap_coordinates[:, components]
        # run lowlevel plot
        return ensemble_crosstab_scatter_lowlevel(
            dataset, component_labels, **kwargs
        )


def projection_scatter_umap_original(
    umap_object: UmapTypeObject,
    color_by: Sequence[float] = None,
    components: Sequence[int] = (0, 1),
    **kwargs,
):
    """
    See `projection_scatter_umap` for more documentation.

    Wrapper around umap.plot.points. Similar to `projection_scatter_umap`, but
    uses the original plotting routine from umap.plot.

    Parameters
    ----------
    umap_object:
        umap object (returned as second output by get_umap_representation)
        to be shown as scatter plot
    color_by:
        A sequence/list of floats, which specify the color in the colormap
    components:
        Components to be plotted (corresponds to columns of umap_coordinates)

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    import umap.plot

    # reduce, if necessary
    umap_object.embedding_ = umap_object.embedding_[:, components]

    # use umap's original plotting routine to visualize
    umap.plot.points(umap_object, values=color_by, theme='viridis', **kwargs)


def projection_scatter_pca(
    pca_coordinates: np.ndarray, components: Sequence[int] = (0, 1), **kwargs
):
    """
    Plot a scatter plot for PCA coordinates.

    Creates either one or multiple scatter plots, depending on the number of
    coordinates passed to it.

    Parameters
    ----------
    pca_coordinates:
        array of pca coordinates (returned as first output by the routine
        get_pca_representation) to be shown as scatter plot
    components:
        Components to be plotted (corresponds to columns of pca_coordinates)

    Returns
    -------
    axs:
        Either one axes object, or a dictionary of plot axes (depending on the
        number of coordinates passed)
    """
    n_components = len(components)
    if n_components == 2:
        # handle components
        x_label = f'PCA component {components[0] + 1}'
        y_label = f'PCA component {components[1] + 1}'

        dataset = pca_coordinates[:, components]

        # call lowlevel routine
        return ensemble_scatter_lowlevel(
            dataset, x_label=x_label, y_label=y_label, **kwargs
        )
    else:
        # We got more than two components. Plot a cross-classification table
        # Create the labels first
        component_labels = [
            f'PCA component {components[i_comp] + 1}'
            for i_comp in range(n_components)
        ]
        # reduce pca components
        dataset = pca_coordinates[:, components]
        # run lowlevel plot
        return ensemble_crosstab_scatter_lowlevel(
            dataset, component_labels, **kwargs
        )


def ensemble_crosstab_scatter_lowlevel(
    dataset: np.ndarray, component_labels: Sequence[str] = None, **kwargs
):
    """
    Plot cross-classification table of scatter plots for different coordinates.

    Lowlevel routine for multiple UMAP and PCA plots, but can also be used to
    visualize, e.g., parameter traces across optimizer runs.

    Parameters
    ----------
    dataset:
        array of data points to be shown as scatter plot
    component_labels:
        labels for the x-axes and the y-axes

    Returns
    -------
    axs:
        A dictionary of plot axes.
    """
    # We got more than two components. Create a cross-classification table
    n_components = dataset.shape[1]
    axs = _create_crosstab_axes(n_components)

    # wo don't even try to plot this into an existing axes object.
    # Overplotting a multi-axes figure is asking for trouble...
    if 'ax' in kwargs.keys():
        del kwargs['ax']

    for x_comp in range(0, n_components - 1):
        for y_comp in range(x_comp + 1, n_components):
            # handle axis labels
            x_label = ''
            y_label = ''
            if x_comp == 0:
                y_label = component_labels[y_comp]
            if y_comp == n_components - 1:
                x_label = component_labels[x_comp]

            # extract the wanted columns
            tmp_dataset = dataset[:, [x_comp, y_comp]]

            # call lowlevel routine
            ensemble_scatter_lowlevel(
                tmp_dataset,
                x_label=x_label,
                y_label=y_label,
                ax=axs[(x_comp, y_comp)],
                **kwargs,
            )
    # return dict of axes
    return axs


def ensemble_scatter_lowlevel(
    dataset,
    ax: Optional[plt.Axes] = None,
    size: Optional[Tuple[float]] = (12, 6),
    x_label: str = 'component 1',
    y_label: str = 'component 2',
    color_by: Sequence[float] = None,
    color_map: str = 'viridis',
    background_color: RGBA = (0.0, 0.0, 0.0, 1.0),
    marker_type: str = '.',
    scatter_size: float = 0.5,
    invert_scatter_order: bool = False,
):
    """
    Create a scatter plot.

    Parameters
    ----------
    dataset:
        array of data points in reduced dimension
    ax:
        Axes object to use.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    x_label:
        The x-axis label
    y_label:
        The y-axis label
    color_by:
        A sequence/list of floats, which specify the color in the colormap
    color_map:
        A colormap name known to pyplot
    background_color:
        Background color of the axes object (defaults to black)
    marker_type:
        Type of plotted markers
    scatter_size:
        Size of plotted markers
    invert_scatter_order:
        Specifies the order of plotting the scatter points, can be important
        in case of overplotting

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # first get the data to check identifiability
    # axes
    if ax is None:
        fig, ax = plt.subplots()
        fig.set_size_inches(*size)
    plt.sca(ax)

    if color_by is None:
        color_by = np.array([1.0] * dataset.shape[0])

    ordering = 1
    if invert_scatter_order:
        ordering = -1

    plt.scatter(
        dataset[::ordering, 0],
        dataset[::ordering, 1],
        c=color_by,
        cmap=color_map,
        marker=marker_type,
        s=scatter_size,
    )

    # beautify
    ax.set_facecolor(background_color)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.xticks([])
    plt.yticks([])

    plt.tight_layout()

    return ax


def _create_crosstab_axes(n_comp: int):
    """
    Create a figure with cross-classification table of axes.

    Parameters
    ----------
    n_comp:
        number of component to be mutually compared

    Returns
    -------
    axs:
        A dictionary of plot axes.
    """
    axs = {}

    # run over x- and y-coordinate
    for x_comp in range(0, n_comp - 1):
        for y_comp in range(x_comp + 1, n_comp):
            i_ax = (y_comp - 1) * (n_comp - 1) + x_comp + 1
            axs[(x_comp, y_comp)] = plt.subplot(n_comp - 1, n_comp - 1, i_ax)

    return axs



"""Inner optimization problem in hierarchical optimization."""
import logging
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd

from ..C import PARAMETER_TYPE, InnerParameterType
from .parameter import InnerParameter

try:
    import amici
    import petab
    from petab.C import (
        ESTIMATE,
        LOWER_BOUND,
        NOISE_PARAMETERS,
        OBSERVABLE_ID,
        OBSERVABLE_PARAMETERS,
        PARAMETER_ID,
        PARAMETER_SCALE,
        TIME,
        UPPER_BOUND,
    )
except ImportError:
    pass

logger = logging.getLogger(__name__)


class InnerProblem:
    """
    Inner optimization problem in hierarchical optimization.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (`num_timepoints` x `num_observables`)
        per simulation condition. Missing observations as NaN.
    """

    def __init__(self, xs: List[InnerParameter], data: List[np.ndarray]):
        self.xs: Dict[str, InnerParameter] = {
            x.inner_parameter_id: x for x in xs
        }
        self.data = data

        logger.debug(f"Created InnerProblem with ids {self.get_x_ids()}")

        if self.is_empty():
            raise ValueError(
                'There are no parameters in the inner problem of hierarchical '
                'optimization.'
            )

    @staticmethod
    def from_petab_amici(
        petab_problem: 'petab.Problem',
        amici_model: 'amici.Model',
        edatas: List['amici.ExpData'],
    ):
        """Create an InnerProblem from a PEtab problem and AMICI objects."""
        return inner_problem_from_petab_problem(
            petab_problem, amici_model, edatas
        )

    def get_x_ids(self) -> List[str]:
        """Get IDs of inner parameters."""
        return list(self.xs.keys())

    def get_xs_for_type(
        self, inner_parameter_type: str
    ) -> List[InnerParameter]:
        """Get inner parameters of the given type."""
        return [
            x
            for x in self.xs.values()
            if x.inner_parameter_type == inner_parameter_type
        ]

    def get_dummy_values(self, scaled: bool) -> Dict[str, float]:
        """
        Get dummy parameter values.

        Get parameter values to be used for simulation before their optimal
        values are computed.

        Parameters
        ----------
        scaled:
            Whether the parameters should be returned on parameter or linear
            scale.
        """
        return {
            x.inner_parameter_id: scale_value(x.dummy_value, x.scale)
            if scaled
            else x.dummy_value
            for x in self.xs.values()
        }

    def get_for_id(self, inner_parameter_id: str) -> InnerParameter:
        """Get InnerParameter for the given parameter ID."""
        try:
            return self.xs[inner_parameter_id]
        except KeyError:
            raise KeyError(f"Cannot find parameter with id {id}.")

    def is_empty(self) -> bool:
        """Check for emptiness.

        Returns
        -------
        ``True`` if there aren't any parameters associated with this problem,
        ``False`` otherwise.
        """
        return len(self.xs) == 0


class AmiciInnerProblem(InnerProblem):
    """
    Inner optimization problem in hierarchical optimization.

    For use with AMICI objects.

    Attributes
    ----------
    edataviews:
        AMICI ``ExpDataView``s for each simulation condition.
    """

    def __init__(self, edatas: List[amici.ExpData], **kwargs):
        super().__init__(**kwargs)

    def check_edatas(self, edatas: List[amici.ExpData]) -> bool:
        """Check for consistency in data.

        Currently only checks for the actual data values. e.g., timepoints are
        not compared.

        Parameters
        ----------
        edatas:
            A data set. Will be checked against the data set provided to the
            constructor.

        Returns
        -------
        Whether the data sets are consistent.
        """
        # TODO replace but edata1==edata2 once this makes it into amici
        #  https://github.com/AMICI-dev/AMICI/issues/1880
        data = [
            amici.numpy.ExpDataView(edata)['observedData'] for edata in edatas
        ]

        if len(self.data) != len(data):
            return False

        for data0, data1 in zip(self.data, data):
            if not np.array_equal(data0, data1, equal_nan=True):
                return False

        return True


def scale_value_dict(
    dct: Dict[str, float], problem: InnerProblem
) -> Dict[str, float]:
    """Scale a value dictionary."""
    scaled_dct = {}
    for key, val in dct.items():
        x = problem.get_for_id(key)
        scaled_dct[key] = scale_value(val, x.scale)
    return scaled_dct


def scale_value(
    val: Union[float, np.array], scale: str
) -> Union[float, np.array]:
    """Scale a single value."""
    if scale == 'lin':
        return val
    if scale == 'log':
        return np.log(val)
    if scale == 'log10':
        return np.log10(val)
    raise ValueError(f"Scale {scale} not recognized.")


def inner_problem_from_petab_problem(
    petab_problem: 'petab.Problem',
    amici_model: 'amici.Model',
    edatas: List['amici.ExpData'],
) -> InnerProblem:
    """
    Create inner problem from PEtab problem.

    Hierarchical optimization is a pypesto-specific PEtab extension.
    """
    import amici

    # inner parameters
    inner_parameters = inner_parameters_from_parameter_df(
        petab_problem.parameter_df
    )

    x_ids = [x.inner_parameter_id for x in inner_parameters]

    # used indices for all measurement specific parameters
    ixs = ixs_for_measurement_specific_parameters(
        petab_problem, amici_model, x_ids
    )

    # transform experimental data
    data = [amici.numpy.ExpDataView(edata)['observedData'] for edata in edatas]

    # matrixify
    ix_matrices = ix_matrices_from_arrays(ixs, data)

    # assign matrices
    for par in inner_parameters:
        par.ixs = ix_matrices[par.inner_parameter_id]

    par_group_types = {
        tuple(obs_pars.split(';')): {
            petab_problem.parameter_df.loc[obs_par, PARAMETER_TYPE]
            for obs_par in obs_pars.split(';')
        }
        for (obs_id, obs_pars), _ in petab_problem.measurement_df.groupby(
            [petab.OBSERVABLE_ID, petab.OBSERVABLE_PARAMETERS], dropna=True
        )
        if ';' in obs_pars  # prefilter for at least 2 observable parameters
    }

    coupled_pars = {
        par
        for group, types in par_group_types.items()
        if (
            (InnerParameterType.SCALING in types)
            and (InnerParameterType.OFFSET in types)
        )
        for par in group
    }

    for par in inner_parameters:
        if par.inner_parameter_type not in [
            InnerParameterType.SCALING,
            InnerParameterType.OFFSET,
        ]:
            continue
        if par.inner_parameter_id in coupled_pars:
            par.coupled = True

    return AmiciInnerProblem(xs=inner_parameters, data=data, edatas=edatas)


def inner_parameters_from_parameter_df(
    df: pd.DataFrame,
) -> List[InnerParameter]:
    """
    Create list of inner free parameters from PEtab parameter table.

    Inner parameters are those that have a non-empty `parameterType` in the
    PEtab problem.
    """
    # create list of hierarchical parameters
    df = df.reset_index()

    for col in (PARAMETER_TYPE,):
        if col not in df:
            df[col] = None

    parameters = []

    for _, row in df.iterrows():
        if not row[ESTIMATE]:
            continue
        if petab.is_empty(row[PARAMETER_TYPE]):
            continue
        parameters.append(
            InnerParameter(
                inner_parameter_id=row[PARAMETER_ID],
                inner_parameter_type=row[PARAMETER_TYPE],
                scale=row[PARAMETER_SCALE],
                lb=row[LOWER_BOUND],
                ub=row[UPPER_BOUND],
            )
        )

    return parameters


def ixs_for_measurement_specific_parameters(
    petab_problem: 'petab.Problem',
    amici_model: 'amici.Model',
    x_ids: List[str],
) -> Dict[str, List[Tuple[int, int, int]]]:
    """
    Create mapping of parameters to measurements.

    Returns
    -------
    A dictionary mapping parameter ID to a List of
    `(condition index, time index, observable index)` tuples in which this
    output parameter is used. For each condition, the time index refers to
    a sorted list of non-unique time points for which there are measurements.
    """
    ixs_for_par = {}
    observable_ids = amici_model.getObservableIds()

    simulation_conditions = (
        petab_problem.get_simulation_conditions_from_measurement_df()
    )
    for condition_ix, condition in simulation_conditions.iterrows():
        # measurement table for current condition
        df_for_condition = petab.get_rows_for_condition(
            measurement_df=petab_problem.measurement_df, condition=condition
        )

        # unique sorted list of timepoints
        timepoints = sorted(df_for_condition[TIME].unique().astype(float))
        # non-unique sorted list of timepoints
        timepoints_w_reps = _get_timepoints_with_replicates(
            measurement_df=df_for_condition
        )

        for time in timepoints:
            # subselect measurements for time `time`
            df_for_time = df_for_condition[df_for_condition[TIME] == time]
            time_ix_0 = timepoints_w_reps.index(time)

            # remember used time indices for each observable
            time_ix_for_obs_ix = {}

            # iterate over measurements
            for _, measurement in df_for_time.iterrows():
                # extract observable index
                observable_ix = observable_ids.index(
                    measurement[OBSERVABLE_ID]
                )

                # as the time indices have to account for replicates, we need
                #  to track which time indices have already been assigned for
                #  the current observable
                if observable_ix in time_ix_for_obs_ix:
                    # a replicate
                    time_ix_for_obs_ix[observable_ix] += 1
                else:
                    # the first measurement for this `(observable, timepoint)`
                    time_ix_for_obs_ix[observable_ix] = time_ix_0
                time_w_reps_ix = time_ix_for_obs_ix[observable_ix]

                observable_overrides = petab.split_parameter_replacement_list(
                    measurement.get(OBSERVABLE_PARAMETERS, None)
                )
                noise_overrides = petab.split_parameter_replacement_list(
                    measurement.get(NOISE_PARAMETERS, None)
                )

                # try to insert if hierarchical parameter
                for override in observable_overrides + noise_overrides:
                    if override in x_ids:
                        ixs_for_par.setdefault(override, []).append(
                            (condition_ix, time_w_reps_ix, observable_ix)
                        )
    return ixs_for_par


def ix_matrices_from_arrays(
    ixs: Dict[str, List[Tuple[int, int, int]]], edatas: List[np.array]
) -> Dict[str, List[np.array]]:
    """
    Convert mapping of parameters to measurements to matrix form.

    Returns
    -------
    A dictionary mapping parameter ID to a list of Boolean matrices, one per
    simulation condition. Therein, ``True`` indicates that the respective
    parameter is used for the model output at the respective timepoint,
    observable and condition index.
    """
    ix_matrices = {
        id: [np.zeros_like(edata, dtype=bool) for edata in edatas]
        for id in ixs
    }
    for id, arr in ixs.items():
        matrices = ix_matrices[id]
        for (condition_ix, time_ix, observable_ix) in arr:
            matrices[condition_ix][time_ix, observable_ix] = True
    return ix_matrices


def _get_timepoints_with_replicates(
    measurement_df: pd.DataFrame,
) -> List[float]:
    """
    Get list of timepoints including replicate measurements.

    :param measurement_df:
        PEtab measurement table subset for a single condition.

    :return:
        Sorted list of timepoints, including multiple timepoints accounting
        for replicate measurements.
    """
    # create sorted list of all timepoints for which measurements exist
    timepoints = sorted(measurement_df[TIME].unique().astype(float))

    # find replicate numbers of time points
    timepoints_w_reps = []
    for time in timepoints:
        # subselect for time
        df_for_time = measurement_df[measurement_df.time == time]
        # rep number is maximum over rep numbers for observables
        n_reps = max(df_for_time.groupby([OBSERVABLE_ID, TIME]).size())
        # append time point n_rep times
        timepoints_w_reps.extend([time] * n_reps)

    return timepoints_w_reps



"""Uniform sampling."""

import numpy as np

from .base import FunctionStartpoints
from .util import rescale


def uniform(
    n_starts: int,
    lb: np.ndarray,
    ub: np.ndarray,
) -> np.ndarray:
    """Generate uniform points.

    Parameters
    ----------
    n_starts: Number of starts.
    lb: Lower bound.
    ub: Upper bound.

    Returns
    -------
    xs: Uniformly sampled points in [lb, ub], shape (n_starts, n_x).
    """
    if not np.isfinite(ub).all() or not np.isfinite(lb).all():
        raise ValueError(
            "Cannot use uniform startpoint method with non-finite boundaries.",
        )

    # parse
    dim = lb.size
    lb = lb.reshape((1, -1))
    ub = ub.reshape((1, -1))

    # create uniform points in [0, 1]
    xs = np.random.random((n_starts, dim))

    # re-scale
    xs = rescale(xs, lb, ub)

    return xs


class UniformStartpoints(FunctionStartpoints):
    """Generate uniformly sampled startpoints."""

    def __init__(
        self,
        use_guesses: bool = True,
        check_fval: bool = False,
        check_grad: bool = False,
    ):
        super().__init__(
            function=uniform,
            use_guesses=use_guesses,
            check_fval=check_fval,
            check_grad=check_grad,
        )



"""Load/reconstitute results from history."""

import logging
import os
from pathlib import Path
from typing import Optional

import h5py
import numpy as np

import pypesto

from ..C import (
    FVAL,
    GRAD,
    HESS,
    HISTORY,
    RES,
    SRES,
    SUFFIXES_CSV,
    SUFFIXES_HDF5,
    TRACE,
    X,
)
from ..history import (
    CsvHistory,
    Hdf5History,
    HistoryOptions,
    HistoryTypeError,
    OptimizerHistory,
)
from ..problem import Problem
from ..result import OptimizeResult, OptimizerResult, Result
from .options import OptimizeOptions

logger = logging.getLogger(__name__)

EXITFLAG_LOADED_FROM_FILE = -99


def fill_result_from_history(
    result: OptimizerResult,
    optimizer_history: OptimizerHistory,
    optimize_options: OptimizeOptions = None,
) -> OptimizerResult:
    """Overwrite some values in the result object with values in the history.

    Parameters
    ----------
    result: Result as reported from the used optimizer.
    optimizer_history: History of function values recorded by the objective.
    optimize_options: Options on e.g. how to override.

    Returns
    -------
    result: The in-place modified result.
    """
    if optimize_options is None:
        optimize_options = OptimizeOptions()

    # logging
    #  function values
    history_fval, result_fval = optimizer_history.fval_min, result.fval
    fval_exist = history_fval is not None and result_fval is not None
    fval_match = fval_exist and np.isclose(history_fval, result_fval)
    if fval_exist and not fval_match:
        logger.debug(
            "Minimal function value mismatch: "
            f"history {history_fval:8e}, result {result_fval:8e}"
        )
    #  parameters
    history_x, result_x = optimizer_history.x_min, result.x
    x_exist = history_x is not None and result_x is not None
    x_match = x_exist and np.allclose(history_x, result_x)
    if x_exist and not x_match:
        logger.debug(
            f"Optimal parameter mismatch: history {history_x}, "
            f"result {result_x}"
        )

    # counters
    # we only use our own counters here as optimizers may report differently
    for key in (FVAL, GRAD, HESS, RES, SRES):
        setattr(
            result, f"n_{key}", getattr(optimizer_history.history, f"n_{key}")
        )

    # initial values
    result.x0 = optimizer_history.x0
    result.fval0 = optimizer_history.fval0

    # trace
    result.history = optimizer_history.history

    # if optimizer is trusted, don't overwrite/complement optimal point
    if not optimize_options.history_beats_optimizer:
        return result

    # exit flag and message
    if isinstance(optimizer_history.history, Hdf5History):
        if (message := optimizer_history.history.message) is not None:
            result.message = message
        if (exitflag := optimizer_history.history.exitflag) is not None:
            result.exitflag = exitflag

    # optimal point
    for key in (X, FVAL, GRAD, HESS, RES, SRES):
        hist_val = getattr(optimizer_history, f"{key}_min")
        # replace by history if history has entry, or point does not match
        #  point recorded in result
        if hist_val is not None or not fval_match or not x_match:
            setattr(result, key, hist_val)

    return result


def read_history_from_file(
    problem: Optional[Problem],
    history_options: HistoryOptions,
    identifier: str,
) -> OptimizerHistory:
    """Read history from file.

    Parameters
    ----------
    problem:
        The problem to find optimal parameters for.
        If ``None``, bounds will be assumed to be [-inf, inf] for checking for
        admissible points.
    identifier:
        Multistart id.
    history_options:
        Optimizer history options.
    """
    if history_options.storage_file is None:
        raise ValueError("No history file specified.")

    # evaluate type
    suffix = Path(history_options.storage_file).suffix[1:]

    if suffix in SUFFIXES_CSV:
        history = CsvHistory(
            file=history_options.storage_file.format(id=identifier),
            options=history_options,
            load_from_file=True,
        )
    elif suffix in SUFFIXES_HDF5:
        history = Hdf5History.load(
            id=identifier,
            file=history_options.storage_file.format(id=identifier),
            options=history_options,
        )
    else:
        raise HistoryTypeError(suffix)

    x0 = history.get_x_trace(0)

    if problem:
        lb, ub = problem.lb, problem.ub
    else:
        lb = np.full_like(x0, fill_value=-np.inf)
        ub = np.full_like(x0, fill_value=np.inf)

    return OptimizerHistory(
        history=history,
        x0=x0,
        lb=lb,
        ub=ub,
        generate_from_history=True,
    )


def read_result_from_file(
    problem: Optional[Problem],
    history_options: HistoryOptions,
    identifier: str,
) -> OptimizerResult:
    """Fill an OptimizerResult from history.

    Parameters
    ----------
    problem:
        The problem to find optimal parameters for.
        If ``None``, bounds will be assumed to be [-inf, inf] for checking for
        admissible points.
    identifier:
        Multistart id.
    history_options:
        Optimizer history options.
    """
    opt_hist = read_history_from_file(
        problem=problem, history_options=history_options, identifier=identifier
    )
    result = OptimizerResult(
        id=identifier,
        message='loaded from file',
        exitflag=EXITFLAG_LOADED_FROM_FILE,
        time=max(opt_hist.history.get_time_trace())
        if len(opt_hist.history)
        else 0.0,
    )
    result.id = identifier
    result = fill_result_from_history(
        result=result,
        optimizer_history=opt_hist,
    )
    if problem:
        result.update_to_full(problem)
    return result


def read_results_from_file(
    problem: Problem,
    history_options: HistoryOptions,
    n_starts: int,
) -> Result:
    """Fill a Result from a set of histories.

    Parameters
    ----------
    problem:
        The problem to find optimal parameters for.
    n_starts:
        Number of performed multistarts.
    history_options:
        Optimizer history options.
    """
    if history_options.storage_file is None:
        raise ValueError("No history file specified.")

    result = Result()
    result.problem = problem
    result.optimize_result = OptimizeResult()
    result.optimize_result.list = [
        read_result_from_file(problem, history_options, str(istart))
        for istart in range(n_starts)
        if os.path.exists(history_options.storage_file.format(id=str(istart)))
    ]
    if not result.optimize_result.list:
        logger.error("No history files found.")

    if len(result.optimize_result.list) != n_starts:
        logger.warning(
            f"History files were incomplete "
            f"({len(result.optimize_result.list)}/{n_starts})."
        )

    result.optimize_result.sort()
    return result


def optimization_result_from_history(
    filename: str,
    problem: pypesto.Problem,
) -> Result:
    """Convert a saved hdf5 History to an optimization result.

    Used for interrupted optimization runs.

    Parameters
    ----------
    filename:
        The name of the file in which the information are stored.
    problem:
        Problem, needed to identify what parameters to accept.

    Returns
    -------
        A result object in which the optimization result is constructed from
        history. But missing "Time, Message and Exitflag" keys.
    """
    result = Result()
    with h5py.File(filename, 'r') as f:
        ids = list(f[HISTORY].keys())
        x0s = [f[f'{HISTORY}/{id}/{TRACE}/0/{X}'][()] for id in ids]

    for id, x0 in zip(ids, x0s):
        history = Hdf5History(id=id, file=filename)
        history.recover_options(filename)
        optimizer_history = OptimizerHistory(
            history=history,
            x0=x0,
            lb=problem.lb,
            ub=problem.ub,
            generate_from_history=True,
        )
        optimizer_result = OptimizerResult(id=id)
        fill_result_from_history(optimizer_result, optimizer_history)
        result.optimize_result.append(optimizer_result)

    return result



"""Functionality related to using a PEtab Select model selection method."""
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import petab_select
from petab_select import (
    VIRTUAL_INITIAL_MODEL,
    CandidateSpace,
    Criterion,
    Method,
    Model,
)

from ..C import TYPE_POSTPROCESSOR
from ..problem import Problem
from .model_problem import ModelProblem


class MethodSignalProceed(str, Enum):
    """Indicators for how a model selection method should proceed."""

    # TODO move to PEtab Select?
    STOP = 'stop'
    CONTINUE = 'continue'


@dataclass
class MethodSignal:
    """The state of a model selection method after a single model calibration.

    Attributes
    ----------
    accept:
        Whether to accept the model.
    proceed:
        How the method should proceed.
    """

    accept: bool
    # TODO change to bool?
    proceed: MethodSignalProceed


class MethodLogger:
    """Log results from a model selection method.

    Attributes
    ----------
    column_width:
        The width of columns when logging.
    column_sep:
        The substring used to separate column values when logging.
    level:
        The logging level.
    logger:
        A logger from the `logging` module.
    """

    column_width: int = 12
    column_sep: str = " | "

    def __init__(self, level: str = 'info'):
        self.logger = logging.getLogger(__name__)
        self.level = level

    def log(self, message, level: str = None) -> None:
        """Log a message.

        Parameters
        ----------
        message:
            The message.
        level:
            The logging level. Defaults to the value defined in the
            constructor.
        """
        if level is None:
            level = self.level
        getattr(self.logger, level)(message)

    def new_selection(self) -> None:
        """Start logging a new model selection."""
        padding = 20
        self.log('-' * padding + 'New Selection' + '-' * padding)
        columns = {
            "Predecessor model subspace:ID": "model0",
            "Model subspace:ID": "model",
            "Criterion ID": "crit",
            "Predecessor model criterion": "model0_crit",
            "Model criterion": "model_crit",
            "Criterion difference": "crit_diff",
            "Accept": "accept",
        }
        columns = {
            k: v.ljust(self.column_width)[: self.column_width]
            for k, v in columns.items()
        }
        self.log(self.column_sep.join(columns.values()))

    def new_result(
        self,
        accept,
        criterion,
        model,
        predecessor_model,
        max_id_length: str = 12,
        precision: int = 3,
    ) -> None:
        """Log a model calibration result.

        Parameters
        ----------
        accept:
            Whether the model is accepted.
        criterion:
            The criterion type.
        max_id_length:
            Model and predecessor model IDs are truncated to this length in the
            logged message.
        model:
            The calibrated model.
        predecessor_model:
            The predecessor model.
        precision:
            The number of decimal places to log.
        """
        model_criterion = model.get_criterion(criterion)

        def get_model_id(model: Model) -> str:
            """Get a model ID for logging.

            Parameters
            ----------
            model:
                The model.

            Returns
            -------
            str
                The ID.
            """
            model_subspace_id = model.model_subspace_id or ''
            original_model_id = model.model_id or model.get_hash()
            model_id = model_subspace_id + ':' + original_model_id
            return model_id

        def float_to_str(value: float, precision: int = 3) -> str:
            return f"{value:.{precision}e}"

        if isinstance(predecessor_model, Model):
            predecessor_model_id = get_model_id(predecessor_model)
            predecessor_model_criterion = predecessor_model.get_criterion(
                criterion
            )
            criterion_difference = float_to_str(
                model_criterion - predecessor_model_criterion
            )
            predecessor_model_criterion = float_to_str(
                predecessor_model_criterion
            )
        else:
            criterion_difference = None
            predecessor_model_criterion = None
            predecessor_model_id = predecessor_model

        model_criterion = float_to_str(model_criterion)

        message_parts = [
            predecessor_model_id,
            get_model_id(model),
            criterion.value,
            predecessor_model_criterion,
            model_criterion,
            criterion_difference,
            accept,
        ]
        message = self.column_sep.join(
            [
                str(v).ljust(self.column_width)[: self.column_width]
                for v in message_parts
            ]
        )
        self.log(message)


class MethodCaller:
    """Handle calls to PEtab Select model selection methods.

    Attributes
    ----------
    petab_select_problem:
        The PEtab Select problem.
    candidate_space:
        A `petab_select.CandidateSpace`, used to generate candidate models.
    criterion:
        The criterion by which models will be compared.
    criterion_threshold:
        The minimum improvement in criterion that a test model must have to
        be selected. The comparison is made according to the method. For
        example, in `ForwardSelector`, test models are compared to the
        previously selected model.
    calibrated_models:
        The calibrated models of the model selection, as a `dict` where keys
        are model hashes and values are models.
    limit:
        Limit the number of calibrated models. NB: the number of accepted
        models may (likely) be fewer.
    logger:
        A `MethodLogger`, used to log results.
    minimize_options:
        A dictionary that will be passed to `pypesto.minimize` as keyword
        arguments for model optimization.
    model_postprocessor:
        A method that is applied to each model after calibration.
    objective_customizer:
        A method that is applied to the pyPESTO objective after the
        objective is initialized, before calibration.
    predecessor_model:
        Specify the predecessor (initial) model for the model selection
        algorithm. If `None`, then the algorithm will generate an
        predecessor model if required.
    select_first_improvement:
        If `True`, model selection will terminate as soon as a better model
        is found. If `False`, all candidate models will be tested.
    startpoint_latest_mle:
        If `True`, one of the startpoints in the multistart optimization
        will be the MLE of the latest model.
    """

    def __init__(
        self,
        petab_select_problem: petab_select.Problem,
        calibrated_models: Dict[str, Model],
        # Arguments/attributes that can simply take the default value here.
        criterion_threshold: float = 0.0,
        limit: int = np.inf,
        minimize_options: Dict = None,
        model_postprocessor: TYPE_POSTPROCESSOR = None,
        objective_customizer: Callable = None,
        select_first_improvement: bool = False,
        startpoint_latest_mle: bool = True,
        # Arguments/attributes that should be handled more carefully.
        candidate_space: CandidateSpace = None,
        criterion: Criterion = None,
        # TODO misleading, `Method` here is simply an Enum, not a callable...
        method: Method = None,
        predecessor_model: Model = None,
        model_to_pypesto_problem_method: Callable[[Any], Problem] = None,
    ):
        """Arguments are used in every `__call__`, unless overridden."""
        self.petab_select_problem = petab_select_problem
        self.calibrated_models = calibrated_models

        self.criterion_threshold = criterion_threshold
        self.limit = limit
        self.minimize_options = minimize_options
        self.model_postprocessor = model_postprocessor
        self.objective_customizer = objective_customizer
        self.predecessor_model = predecessor_model
        self.select_first_improvement = select_first_improvement
        self.startpoint_latest_mle = startpoint_latest_mle
        self.model_to_pypesto_problem_method = model_to_pypesto_problem_method

        self.logger = MethodLogger()

        self.criterion = criterion
        if self.criterion is None:
            self.criterion = self.petab_select_problem.criterion

        # Forbid specification of both a candidate space and a method.
        if candidate_space is not None and method is not None:
            self.logger.log(
                (
                    'Both `candidate_space` and `method` were provided. '
                    'Please only provide one. The method will be ignored here.'
                ),
                level='warning',
            )
        # Get method.
        self.method = (
            method
            if method is not None
            else candidate_space.method
            if candidate_space is not None
            else self.petab_select_problem.method
        )
        # Require either a candidate space or a method.
        if candidate_space is None and self.method is None:
            raise ValueError(
                'Please provide one of either `candidate_space` or `method`, '
                'or specify the `method` in the PEtab Select problem.'
            )
        # Use candidate space if provided.
        if candidate_space is not None:
            self.candidate_space = candidate_space
            if predecessor_model is not None:
                candidate_space.set_predecessor_model(predecessor_model)
        # Else generate one based on the PEtab Select problem.
        else:
            self.candidate_space = (
                self.petab_select_problem.new_candidate_space(
                    method=self.method,
                    predecessor_model=self.predecessor_model,
                )
            )
        # May have changed from `None` to `petab_select.VIRTUAL_INITIAL_MODEL`
        self.predecessor_model = self.candidate_space.get_predecessor_model()

    def __call__(
        self,
        predecessor_model: Optional[Union[Model, None]] = None,
        newly_calibrated_models: Optional[Dict[str, Model]] = None,
    ) -> Tuple[List[Model], Dict[str, Model]]:
        """Run a single iteration of the model selection method.

        A single iteration here refers to calibration of all candidate models.
        For example, given a predecessor model with 3 estimated parameters,
        with the forward method, a single iteration would involve calibration
        of all models that have both: the same 3 estimated parameters; and 1
        additional estimated paramenter.

        The input `newly_calibrated_models` is from the previous iteration. The
        output `newly_calibrated_models` is from the current iteration.

        Parameters
        ----------
        predecessor_model:
            The model that will be used for comparison. Example 1: the
            initial model of a forward method. Example 2: all models found
            with a brute force method should be better than this model.
        newly_calibrated_models:
            The newly calibrated models from the previous iteration.

        Returns
        -------
        tuple
            A 2-tuple, with the following values:

               1. the predecessor model for the newly calibrated models; and
               2. the newly calibrated models, as a `dict` where keys are model
                  hashes and values are models.
        """
        # All calibrated models in this iteration (see second return value).
        self.logger.new_selection()

        if predecessor_model is None:
            # May still be `None` (e.g. brute force method)
            predecessor_model = self.predecessor_model

        candidate_space = petab_select.ui.candidates(
            problem=self.petab_select_problem,
            candidate_space=self.candidate_space,
            limit=self.limit,
            calibrated_models=self.calibrated_models,
            newly_calibrated_models=newly_calibrated_models,
            excluded_model_hashes=self.calibrated_models.keys(),
            # FIXME meaning changes here, i.e. predecessor model is used as previous
            #       predecessor model, since PEtab Select now gets the next predecessor
            #       model if the candidate space has models.
            previous_predecessor_model=predecessor_model,
            criterion=self.criterion,
        )
        predecessor_model = self.candidate_space.predecessor_model

        if not candidate_space.models:
            raise StopIteration("No valid models found.")

        # TODO parallelize calibration (maybe not sensible if
        #      `self.select_first_improvement`)
        newly_calibrated_models = {}
        for candidate_model in candidate_space.models:
            # autoruns calibration
            self.new_model_problem(model=candidate_model)
            newly_calibrated_models[
                candidate_model.get_hash()
            ] = candidate_model
            method_signal = self.handle_calibrated_model(
                model=candidate_model,
                predecessor_model=predecessor_model,
            )
            if method_signal.proceed == MethodSignalProceed.STOP:
                break

        self.calibrated_models.update(newly_calibrated_models)

        return predecessor_model, newly_calibrated_models

    def handle_calibrated_model(
        self,
        model: Model,
        predecessor_model: Optional[Model],
    ) -> MethodSignal:
        """Handle the model selection method, given a new calibrated model.

        Parameters
        ----------
        model:
            The calibrated model.
        predecessor_model:
            The predecessor model.

        Returns
        -------
        MethodSignal
            A `MethodSignal` that describes the result.
        """
        # Use the predecessor model from `__init__` if an iteration-specific
        # predecessor model was not supplied to `__call__`.
        if predecessor_model is None:
            # May still be `None` after this assignment.
            predecessor_model = self.predecessor_model

        # Default to accepting the model and continuing the method.
        method_signal = MethodSignal(
            accept=True,
            proceed=MethodSignalProceed.CONTINUE,
        )

        # Reject the model if it doesn't improve on the predecessor model.
        if (
            predecessor_model is not None
            and predecessor_model != VIRTUAL_INITIAL_MODEL
            and not self.model1_gt_model0(
                model1=model, model0=predecessor_model
            )
        ):
            method_signal.accept = False

        # Stop the model selection method if it a first improvement is found.
        if self.select_first_improvement and method_signal.accept:
            method_signal.proceed = MethodSignalProceed.STOP

        # TODO allow users to supply an arbitrary constraint function to e.g.:
        #      - quit after 10 accepted models
        #      - reject models that are worse than the current 10 best models

        # Log result
        self.logger.new_result(
            accept=method_signal.accept,
            criterion=self.criterion,
            model=model,
            predecessor_model=predecessor_model,
        )

        return method_signal

    def model1_gt_model0(
        self,
        model1: Model,
        model0: Model,
    ) -> bool:
        """Compare models by criterion.

        Parameters
        ----------
        model1:
            The new model.
        model0:
            The original model.

        Returns
        -------
        bool
            `True`, if `model1` is superior to `model0` by the criterion,
            else `False`.
        """
        if self.criterion in [
            Criterion.AIC,
            Criterion.AICC,
            Criterion.BIC,
            Criterion.LH,
            Criterion.LLH,
            Criterion.NLLH,
        ]:
            result = petab_select.model.default_compare(
                model0=model0,
                model1=model1,
                criterion=self.criterion,
                criterion_threshold=self.criterion_threshold,
            )
        else:
            raise NotImplementedError(
                f"Model selection criterion: {self.criterion}."
            )
        return result

    def new_model_problem(
        self,
        model: Model,
        valid: bool = True,
        autorun: bool = True,
    ) -> ModelProblem:
        """Create a model problem, usually to calibrate a model.

        Parameters
        ----------
        model:
            The model.
        valid:
            Whether the model should be considered a valid model. If it is
            not valid, it will not be calibrated.
        autorun:
            Whether the model should be calibrated upon creation.

        Returns
        -------
        ModelProblem
            The model selection problem.
        """
        x_guess = None
        if (
            self.startpoint_latest_mle
            and model.predecessor_model_hash in self.calibrated_models
        ):
            predecessor_model = self.calibrated_models[
                model.predecessor_model_hash
            ]
            if str(model.petab_yaml) != str(predecessor_model.petab_yaml):
                raise NotImplementedError(
                    'The PEtab YAML files differ between the model and its '
                    'predecessor model. This may imply different (fixed union '
                    'estimated) parameter sets. Support for this is not yet '
                    'implemented.'
                )
            x_guess = {
                **predecessor_model.parameters,
                **predecessor_model.estimated_parameters,
            }

        return ModelProblem(
            model=model,
            criterion=self.criterion,
            valid=valid,
            autorun=autorun,
            x_guess=x_guess,
            minimize_options=self.minimize_options,
            objective_customizer=self.objective_customizer,
            postprocessor=self.model_postprocessor,
            model_to_pypesto_problem_method=self.model_to_pypesto_problem_method,
        )



from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle

from ..C import COLOR_HIT_BOTH_BOUNDS, COLOR_HIT_NO_BOUNDS, COLOR_HIT_ONE_BOUND
from ..ensemble import Ensemble


def ensemble_identifiability(
    ensemble: Ensemble,
    ax: Optional[plt.Axes] = None,
    size: Optional[Tuple[float]] = (12, 6),
):
    """
    Visualize identifiablity of parameter ensemble.

    Plot an overview about how many parameters hit the parameter bounds based
    on a ensemble of parameters. confidence intervals/credible ranges are
    computed via the ensemble mean plus/minus 1 standard deviation.
    This highlevel routine expects a ensemble object as input.

    Parameters
    ----------
    ensemble:
        ensemble of parameter vectors (from pypesto.ensemble)
    ax:
        Axes object to use.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # first get the data to check identifiability
    id_df = ensemble.check_identifiability()

    # check how many bounds are actually hit and which ones
    none_hit, lb_hit, ub_hit, both_hit = _prepare_identifiability_plot(id_df)

    # call lowlevel routine whick works with np arrays only
    ax = ensemble_identifiability_lowlevel(
        none_hit, lb_hit, ub_hit, both_hit, ax, size
    )

    return ax


def ensemble_identifiability_lowlevel(
    none_hit: np.ndarray,
    lb_hit: np.ndarray,
    ub_hit: np.ndarray,
    both_hit: np.ndarray,
    ax: Optional[plt.Axes] = None,
    size: Optional[Tuple[float]] = (16, 10),
):
    """
    Low-level identifiablity routine.

    Plot an overview about how many parameters hit the parameter bounds based
    on a ensemble of parameters. Confidence intervals/credible ranges are
    computed via the ensemble mean plus/minus 1 standard deviation.
    This lowlevel routine works with numpy arrays which define the confidence
    intervals/credible ranges of each parameter.

    Parameters
    ----------
    none_hit:
        2-dimensional array of confidence interval/credible ranges for
        identifiable parameters
    lb_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the lower parameter bound
    ub_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the upper parameter bound
    both_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit both parameter bounds
    ax:
        Axes object to use.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # define some short hands for later plotting
    n_par = sum(
        [
            none_hit.shape[0],
            lb_hit.shape[0],
            ub_hit.shape[0],
            both_hit.shape[0],
        ]
    )
    x_both = len(both_hit) / n_par
    x_lb = len(lb_hit) / n_par
    x_ub = len(ub_hit) / n_par
    x_none = 1.0 - x_both - x_ub - x_lb

    (
        patches_both_hit,
        patches_lb_hit,
        patches_ub_hit,
        patches_none_hit,
    ) = _create_patches(none_hit, lb_hit, ub_hit, both_hit)

    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # create axes object and add patch collections
    if patches_both_hit:
        ax.add_collection(patches_both_hit)
    if patches_lb_hit:
        ax.add_collection(patches_lb_hit)
    if patches_ub_hit:
        ax.add_collection(patches_ub_hit)
    if patches_none_hit:
        ax.add_collection(patches_none_hit)

    # plot dashed lines indicating the number rof non-identifiable parameters
    vert = [-0.05, 1.05]
    ax.plot([x_both, x_both], vert, 'k--', linewidth=1.5)
    ax.plot([x_both + x_lb, x_both + x_lb], vert, 'k--', linewidth=1.5)
    ax.plot(
        [x_both + x_lb + x_ub, x_both + x_lb + x_ub],
        vert,
        'k--',
        linewidth=1.5,
    )

    # add text
    if patches_both_hit:
        ax.text(
            x_both / 2,
            -0.05,
            'both bounds hit',
            color=COLOR_HIT_BOTH_BOUNDS,
            rotation=-90,
            va='top',
            ha='center',
        )
    if patches_lb_hit:
        ax.text(
            x_both + x_lb / 2,
            -0.05,
            'lower bound hit',
            color=COLOR_HIT_ONE_BOUND,
            rotation=-90,
            va='top',
            ha='center',
        )
    if patches_ub_hit:
        ax.text(
            x_both + x_lb + x_ub / 2,
            -0.05,
            'upper bound hit',
            color=COLOR_HIT_ONE_BOUND,
            rotation=-90,
            va='top',
            ha='center',
        )
    if patches_none_hit:
        ax.text(
            1 - x_none / 2,
            -0.05,
            'no bounds hit',
            color=COLOR_HIT_NO_BOUNDS,
            rotation=-90,
            va='top',
            ha='center',
        )
    ax.text(
        0,
        -0.7,
        'identifiable parameters: {:4.1f}%'.format(x_none * 100),
        va='top',
    )

    # plot upper and lower bounds
    ax.text(-0.03, 1.0, 'upper\nbound', ha='right', va='center')
    ax.text(-0.03, 0.0, 'lower\nbound', ha='right', va='center')
    ax.plot([-0.02, 1.03], [0, 0], 'k:', linewidth=1.5)
    ax.plot([-0.02, 1.03], [1, 1], 'k:', linewidth=1.5)
    plt.xticks([])
    plt.yticks([])

    # plot frame
    ax.plot([0, 0], vert, 'k-', linewidth=1.5)
    ax.plot([1, 1], vert, 'k-', linewidth=1.5)

    # beautify axes
    plt.xlim((-0.15, 1.1))
    plt.ylim((-0.78, 1.15))
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)

    return ax


def _prepare_identifiability_plot(id_df: pd.DataFrame):
    """
    Group model parameters based on an ensemble object .

    Can group into four categories, based on the mean of the parameter
    ensemble plus/minus 1 standard deviation: Parameters that hit both
    bounds, parameters that hit only the lower [or upper] bound,
    and parameters that hit no bounds. It returns them as four numpy arrays,
    together with their confidence intervals/credible ranges.

    Parameters
    ----------
    id_df:
        Pandas dataframe with information about parameter identifiability,
        as created by pypesto.ensemble.check_identifiability()

    Returns
    -------
    none_hit:
        2-dimensional array of confidence interval/credible ranges for
        identifiable parameters

    lb_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the lower parameter bound

    ub_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the upper parameter bound

    both_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit both parameter bounds
    """
    # prepare
    both_hit = []
    lb_hit = []
    ub_hit = []
    none_hit = []

    def _affine_transform(par_info):
        # rescale parameters to bounds
        lb = par_info['lowerBound']
        ub = par_info['upperBound']
        val_l = par_info['ensemble_mean'] - par_info['ensemble_std']
        val_u = par_info['ensemble_mean'] + par_info['ensemble_std']
        # check if parameter confidence intervals/credible ranges hit bound
        if val_l <= lb:
            lower_val = 0.0
        else:
            lower_val = (val_l - lb) / (ub - lb)
        if val_u >= ub:
            upper_val = 1.0
        else:
            upper_val = (val_u - lb) / (ub - lb)

        return lower_val, upper_val

    for par_id in list(id_df.index):
        # check which of the parameters seems to be identifiable and group them
        if (
            id_df.loc[par_id, 'within lb: 1 std']
            and id_df.loc[par_id, 'within ub: 1 std']
        ):
            none_hit.append(_affine_transform(id_df.loc[par_id, :]))
        elif id_df.loc[par_id, 'within lb: 1 std']:
            ub_hit.append(_affine_transform(id_df.loc[par_id, :]))
        elif id_df.loc[par_id, 'within ub: 1 std']:
            lb_hit.append(_affine_transform(id_df.loc[par_id, :]))
        else:
            both_hit.append(_affine_transform(id_df.loc[par_id, :]))

    return (
        np.array(none_hit),
        np.array(lb_hit),
        np.array(ub_hit),
        np.array(both_hit),
    )


def _create_patches(
    none_hit: np.ndarray,
    lb_hit: np.ndarray,
    ub_hit: np.ndarray,
    both_hit: np.ndarray,
):
    """
    Create patches for identifiability analysis.

    Create matplotlib.patches.PatchCollection objects from numpy arrays with
    confidence intervals/credible ranges, which visualize identifiability
    properties of the model parameters.

    Parameters
    ----------
    none_hit:
        2-dimensional array of confidence interval/credible ranges for
        identifiable parameters

    lb_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the lower parameter bound

    ub_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the upper parameter bound

    both_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit both parameter bounds

    Returns
    -------
    patches_both_hit:
        patches showing parameters which hit both parameter bounds in the
        ensemble (and are hence non-identifiable)

    patches_lb_hit:
        patches showing parameters which hit only the lower parameter bounds
        in the ensemble (and are hence non-identifiable)

    patches_ub_hit:
        patches showing parameters which hit only the lower parameter bounds
        in the ensemble (and are hence non-identifiable)

    patches_none_hit
        patches showing parameters which hit no parameter bounds in the
        ensemble (and are hence identifiable)
    """
    # get total number of parameters
    n_par = sum(
        [
            none_hit.shape[0],
            lb_hit.shape[0],
            ub_hit.shape[0],
            both_hit.shape[0],
        ]
    )

    # start patches at the left end and increment by h = 1/n_par
    x = 0.0
    h = 1.0 / n_par

    # creates patches for parameters which hit both bounds
    patches_both_hit = []
    if both_hit.size > 0:
        for _ in both_hit:
            # create a list of rectangles
            patches_both_hit.append(Rectangle((x, 0.0), h, 1.0))
            x += h
        patches_both_hit = PatchCollection(
            patches_both_hit, facecolors=COLOR_HIT_BOTH_BOUNDS
        )

    # creates patches for parameters which hit lower bound
    patches_lb_hit = []
    # sort by normalizes length of confidence interval/credible range
    if lb_hit.size > 0:
        tmp_lb = np.sort(lb_hit[:, 1])[::-1]
        for lb_par in tmp_lb:
            # create a list of rectangles
            patches_lb_hit.append(Rectangle((x, 0.0), h, lb_par))
            x += h
        patches_lb_hit = PatchCollection(
            patches_lb_hit, facecolors=COLOR_HIT_ONE_BOUND
        )

    # creates patches for parameters which hit upper bound
    patches_ub_hit = []
    # sort by normalizes length of confidence interval/credible range
    if ub_hit.size > 0:
        tmp_ub = np.sort(ub_hit[:, 0])
        for ub_par in tmp_ub:
            # create a list of rectangles
            patches_ub_hit.append(Rectangle((x, ub_par), h, 1.0 - ub_par))
            x += h
        patches_ub_hit = PatchCollection(
            patches_ub_hit, facecolors=COLOR_HIT_ONE_BOUND
        )

    # creates patches for parameters which hit no bounds
    patches_none_hit = []
    # sort by normalizes length of confidence interval/credible range
    if none_hit.size > 0:
        tmp_none = np.argsort(none_hit[:, 1] - none_hit[:, 0])[::-1]
        for none_par in tmp_none:
            patches_none_hit.append(
                # create a list of rectangles
                Rectangle(
                    (x, none_hit[none_par, 0]),
                    h,
                    none_hit[none_par, 1] - none_hit[none_par, 0],
                )
            )
            x += h
        patches_none_hit = PatchCollection(
            patches_none_hit, facecolors=COLOR_HIT_NO_BOUNDS
        )

    return patches_both_hit, patches_lb_hit, patches_ub_hit, patches_none_hit



"""Helper for objective evaluation during scatter search."""

import threading
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from typing import Optional, Sequence, Tuple

import numpy as np

from pypesto import Problem
from pypesto.startpoint import StartpointMethod


class FunctionEvaluator:
    """Wrapper for optimization problem and startpoint method.

    Takes care of (potentially parallel) function evaluations, startpoint
    sampling, and tracks number of function evaluations.

    Attributes
    ----------
    problem: The problem
    startpoint_method: Method for choosing feasible parameters
    n_eval: Number of objective evaluations since construction or last call to
        ``reset_counter``.
    n_eval_round: Number of objective evaluations since construction or last
        call to ``reset_counter`` or ``reset_round_counter``.
    """

    def __init__(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
        n_threads: int,
    ):
        """Construct.

        Parameters
        ----------
        problem: The problem
        startpoint_method: Method for choosing feasible parameters
        n_threads: Maximum number of threads to use for parallel objective
            function evaluations. Requires the objective to be copy-able, and
            that copies are thread-safe.
        """
        self.problem: Problem = problem
        self.startpoint_method: StartpointMethod = startpoint_method
        self.n_eval: int = 0
        self.n_eval_round: int = 0
        # Number of threads for parallel objective evaluation
        self._n_threads: int = n_threads

        self._init_threading()

    def __getstate__(self):
        return {
            k: v
            for k, v in vars(self).items()
            if k not in {'_thread_local', '_executor'}
        }

    def __setstate__(self, state):
        vars(self).update(state)
        self._init_threading()

    def _init_threading(self):
        """Initialize multi-threading-related attributes."""
        # Thread-local storage for copies of objective. Each thread gets its
        #  own copy of the objective, which may be sufficient to make some
        #  objectives thread-safe.
        self._thread_local: threading.local = threading.local()
        # The thread-pool to be used for parallel objective evaluations
        self._executor: Optional[ThreadPoolExecutor] = (
            ThreadPoolExecutor(
                max_workers=self._n_threads,
                thread_name_prefix=__name__,
                initializer=self._initialize_worker,
                initargs=(self._thread_local,),
            )
            if self._n_threads > 1
            else None
        )

    def single(self, x: np.array) -> float:
        """Evaluate objective at point ``x``.

        Returns
        -------
        The objective function value at ``x``.
        """
        self.n_eval += 1
        self.n_eval_round += 1
        return self.problem.objective(x)

    def multiple(self, xs: Sequence[np.ndarray]) -> np.array:
        """Evaluate objective at several points.

        Parameters
        ----------
        xs: Sequence of parameter vectors at which the objective is to be
            evaluated.

        Returns
        -------
        The objective function values in the same order as the inputs.
        """
        if self._executor is not None:
            res = np.fromiter(
                self._executor.map(
                    self._evaluate_on_worker,
                    ((self._thread_local, x) for x in xs),
                ),
                dtype=float,
            )
        else:
            res = np.fromiter(map(self.single, xs), dtype=float)
        self.n_eval += len(xs)
        self.n_eval_round += len(xs)
        return res

    def single_random(self) -> Tuple[np.array, float]:
        """Evaluate objective at a random point.

        The point is generated by the given``startpoint_method``. A new point
        will be generated until a finite objective value is obtained.

        Returns
        -------
        Tuple of the generated parameter vector and the respective function
        value.
        """
        x = fx = np.nan
        while not np.isfinite(fx):
            x = self.startpoint_method(n_starts=1, problem=self.problem)[0]
            fx = self.single(x)
        return x, fx

    def multiple_random(self, n: int) -> Tuple[np.array, np.array]:
        """Evaluate objective at ``n`` random points.

        The points are generated by the given``startpoint_method``. New points
        will be generated until only finite objective values are obtained.

        Returns
        -------
        Tuple of the generated parameter vectors and the respective function
        values.
        """
        fxs = np.full(shape=n, fill_value=np.nan)
        xs = np.full(shape=(n, self.problem.dim), fill_value=np.nan)

        while not np.isfinite(fxs).all():
            retry_indices = np.argwhere(~np.isfinite(fxs)).squeeze()
            xs[retry_indices] = self.startpoint_method(
                n_starts=len(retry_indices), problem=self.problem
            )
            fxs[retry_indices] = self.multiple(xs[retry_indices])
        return xs, fxs

    def reset_counter(self):
        """Reset the overall function counter."""
        self.n_eval = 0
        self.reset_round_counter()

    def reset_round_counter(self):
        """Reset the round function counter."""
        self.n_eval_round = 0

    def _initialize_worker(self, local):
        """Initialize worker threads."""
        # create a copy of the objective to maybe be thread-safe.
        local.objective = deepcopy(self.problem.objective)

    @staticmethod
    def _evaluate_on_worker(local_and_x) -> float:
        """Task handler on worker threads."""
        local, x = local_and_x
        return local.objective(x)



import logging
from typing import Any, Literal

import numpy as np

from ..C import (
    DUMMY_INNER_VALUE,
    INNER_PARAMETER_BOUNDS,
    LIN,
    LOG,
    LOG10,
    LOWER_BOUND,
    UPPER_BOUND,
    InnerParameterType,
)

logger = logging.getLogger(__name__)


class InnerParameter:
    """
    An inner parameter of a hierarchical optimization problem.

    Attributes
    ----------
    coupled:
        Whether the inner parameter is part of an observable that has both
        an offset and scaling inner parameter.
    dummy_value:
        Value to be used when the optimal parameter is not yet known
        (in particular to simulate unscaled observables).
    inner_parameter_id:
        The inner parameter ID.
    inner_parameter_type:
        The inner parameter type.
    ixs:
        A mask (boolean matrix) that indicates the measurements that this
        parameter is used in.
    lb:
        The lower bound, for optimization.
    scale:
        Scale on which to estimate this parameter.
    ub:
        The upper bound, for optimization.
    """

    def __init__(
        self,
        inner_parameter_id: str,
        inner_parameter_type: InnerParameterType,
        scale: Literal[LIN, LOG, LOG10] = LIN,
        lb: float = -np.inf,
        ub: float = np.inf,
        ixs: Any = None,
        dummy_value: float = None,
    ):
        """
        Construct.

        Parameters
        ----------
        See class attributes.
        """
        self.inner_parameter_id: str = inner_parameter_id
        self.coupled = False
        self.inner_parameter_type: str = inner_parameter_type

        if scale not in {LIN, LOG, LOG10}:
            raise ValueError(f"Scale not recognized: {scale}.")
        self.scale = scale

        if inner_parameter_type not in (
            InnerParameterType.OFFSET,
            InnerParameterType.SIGMA,
            InnerParameterType.SCALING,
        ):
            raise ValueError(
                f"Unsupported inner parameter type `{inner_parameter_type}`."
            )

        self.lb: float = lb
        self.ub: float = ub
        self.check_bounds()
        self.ixs: Any = ixs

        if dummy_value is None:
            try:
                dummy_value = DUMMY_INNER_VALUE[inner_parameter_type]
            except KeyError as e:
                raise ValueError(
                    "Unsupported parameter type. Parameter id:"
                    f"`{inner_parameter_id}`. Parameter type:"
                    f"`{inner_parameter_type}`."
                ) from e

        self.dummy_value: float = dummy_value

    def check_bounds(self):
        """Check bounds."""
        expected_lb = INNER_PARAMETER_BOUNDS[self.inner_parameter_type][
            LOWER_BOUND
        ]
        expected_ub = INNER_PARAMETER_BOUNDS[self.inner_parameter_type][
            UPPER_BOUND
        ]
        if self.lb != expected_lb or self.ub != expected_ub:
            raise ValueError(
                "Invalid bounds for inner parameters. Parameter ID: "
                f"`{self.inner_parameter_id}`. Provided bounds: "
                f"`[{self.lb}, {self.ub}]`. Expected bounds: "
                f"`[{expected_lb}, {expected_ub}]`. "
                f"All expected parameter bounds:\n{INNER_PARAMETER_BOUNDS}"
            )



"""Overall result."""

from .optimize import OptimizeResult
from .profile import ProfileResult
from .sample import SampleResult


class Result:
    """
    Universal result object for pypesto.

    The algorithms like optimize, profile,
    sample fill different parts of it.

    Attributes
    ----------
    problem: pypesto.Problem
        The problem underlying the results.
    optimize_result:
        The results of the optimizer runs.
    profile_result:
        The results of the profiler run.
    sample_result:
        The results of the sampler run.
    """

    def __init__(self, problem=None):
        self.problem = problem
        self.optimize_result = OptimizeResult()
        self.profile_result = ProfileResult()
        self.sample_result = SampleResult()

    def summary(self) -> str:
        """Get summary of the object."""
        return self.optimize_result.summary()



"""Base history class."""

import numbers
import time
from abc import ABC, abstractmethod
from typing import Dict, Sequence, Tuple, Union

import numpy as np

from ..C import (
    FVAL,
    GRAD,
    HESS,
    MODE_FUN,
    MODE_RES,
    RES,
    SRES,
    TIME,
    ModeType,
    X,
)
from ..util import (
    fval_to_chi2,
    grad_to_schi2,
    res_to_fval,
    sres_to_fim,
    sres_to_grad,
)
from .options import HistoryOptions
from .util import MaybeArray, ResultDict


class HistoryBase(ABC):
    """Abstract base class for histories."""

    # values calculated by the objective function
    RESULT_KEYS = (FVAL, GRAD, HESS, RES, SRES)
    # all possible history entries
    ALL_KEYS = (X, *RESULT_KEYS, TIME)

    def __init__(self, options: HistoryOptions = None):
        if options is None:
            options = HistoryOptions()
        options = HistoryOptions.assert_instance(options)
        self.options: HistoryOptions = options

    @abstractmethod
    def update(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        """
        Update history after a function evaluation.

        Parameters
        ----------
        x:
            The parameter vector.
        sensi_orders:
            The sensitivity orders computed.
        mode:
            The objective function mode computed (function value or residuals).
        result:
            The objective function values for parameters `x`, sensitivities
            `sensi_orders` and mode `mode`.
        """

    def finalize(
        self,
        message: str = None,
        exitflag: str = None,
    ) -> None:
        """
        Finalize history. Called after a run. Default: Do nothing.

        Parameters
        ----------
        message:
            Optimizer message to be saved.
        exitflag:
            Optimizer exitflag to be saved.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Define length by number of stored entries in the history."""

    @property
    @abstractmethod
    def n_fval(self) -> int:
        """Return number of function evaluations."""

    @property
    @abstractmethod
    def n_grad(self) -> int:
        """Return number of gradient evaluations."""

    @property
    @abstractmethod
    def n_hess(self) -> int:
        """Return number of Hessian evaluations."""

    @property
    @abstractmethod
    def n_res(self) -> int:
        """Return number of residual evaluations."""

    @property
    @abstractmethod
    def n_sres(self) -> int:
        """Return number or residual sensitivity evaluations."""

    @property
    @abstractmethod
    def start_time(self) -> float:
        """Return start time."""

    @abstractmethod
    def get_x_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        """
        Return parameters.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """

    @abstractmethod
    def get_fval_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[float], float]:
        """
        Return function values.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """

    @abstractmethod
    def get_grad_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """
        Return gradients.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """

    @abstractmethod
    def get_hess_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """
        Return hessians.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """

    @abstractmethod
    def get_res_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """
        Residuals.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """

    @abstractmethod
    def get_sres_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """
        Residual sensitivities.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """

    def get_chi2_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[float], float]:
        """
        Chi2 values.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        fval_trace = self.get_fval_trace(ix, trim)
        reduce = isinstance(ix, numbers.Integral)
        if reduce:
            return fval_to_chi2(fval_trace)
        else:
            return [fval_to_chi2(fval) for fval in fval_trace]

    def get_schi2_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """
        Chi2 sensitivities.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        grad_trace = self.get_grad_trace(ix, trim)
        reduce = isinstance(ix, numbers.Integral)
        if reduce:
            return grad_to_schi2(grad_trace)
        else:
            return [grad_to_schi2(grad) for grad in grad_trace]

    @abstractmethod
    def get_time_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[float], float]:
        """
        Cumulative execution times.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_trimmed_indices(self) -> np.ndarray:
        """Get indices for a monotonically decreasing history."""
        fval_trace = self.get_fval_trace()
        return np.where(fval_trace <= np.fmin.accumulate(fval_trace))[0]

    def implements_trace(self) -> bool:
        """Check whether the history has a trace that can be queried."""
        try:
            self.get_fval_trace()
        except NotImplementedError:
            return False

        return True


class NoHistory(HistoryBase):
    """Dummy history that does not do anything.

    Can be used whenever a history object is needed, but no history is desired.
    Can be created, but not queried.
    """

    def update(  # noqa: D102
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        pass

    def __len__(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def n_fval(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def n_grad(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def n_hess(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def n_res(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def n_sres(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def start_time(self) -> float:  # noqa: D102
        raise NotImplementedError()

    def get_x_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        raise NotImplementedError()

    def get_fval_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:
        raise NotImplementedError()

    def get_grad_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_hess_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_res_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_sres_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_time_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:
        raise NotImplementedError()


class CountHistoryBase(HistoryBase):
    """Abstract class tracking counts of function evaluations.

    Needs a separate implementation of trace.
    """

    def __init__(self, options: Union[HistoryOptions, Dict] = None):
        super().__init__(options)
        self._n_fval: int = 0
        self._n_grad: int = 0
        self._n_hess: int = 0
        self._n_res: int = 0
        self._n_sres: int = 0
        self._start_time: float = time.time()

    def update(  # noqa: D102
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        self._update_counts(sensi_orders, mode)

    def _update_counts(
        self,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
    ):
        """Update the counters."""
        if mode == MODE_FUN:
            if 0 in sensi_orders:
                self._n_fval += 1
            if 1 in sensi_orders:
                self._n_grad += 1
            if 2 in sensi_orders:
                self._n_hess += 1
        elif mode == MODE_RES:
            if 0 in sensi_orders:
                self._n_res += 1
            if 1 in sensi_orders:
                self._n_sres += 1

    @property
    def n_fval(self) -> int:  # noqa: D102
        return self._n_fval

    @property
    def n_grad(self) -> int:  # noqa: D102
        return self._n_grad

    @property
    def n_hess(self) -> int:  # noqa: D102
        return self._n_hess

    @property
    def n_res(self) -> int:  # noqa: D102
        return self._n_res

    @property
    def n_sres(self) -> int:  # noqa: D102
        return self._n_sres

    @property
    def start_time(self) -> float:  # noqa: D102
        return self._start_time


class CountHistory(CountHistoryBase):
    """History that can only count, other functions cannot be invoked."""

    def __len__(self) -> int:  # noqa: D102
        raise NotImplementedError()

    def get_x_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        raise NotImplementedError()

    def get_fval_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:
        raise NotImplementedError()

    def get_grad_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_hess_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_res_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_sres_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        raise NotImplementedError()

    def get_time_trace(  # noqa: D102
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:

        raise NotImplementedError()


def add_fun_from_res(result: ResultDict) -> ResultDict:
    """Calculate function values from residual values.

    Copies the result, but apart performs calculations only if entries
    are not present yet in the result object
    (thus can be called repeatedly).

    Parameters
    ----------
    result: Result dictionary from the objective function.

    Returns
    -------
    full_result:
        Result dicionary, adding whatever is possible to calculate.
    """
    result = result.copy()

    # calculate function values from residuals
    if result.get(FVAL) is None:
        result[FVAL] = res_to_fval(result.get(RES))
    if result.get(GRAD) is None:
        result[GRAD] = sres_to_grad(result.get(RES), result.get(SRES))
    if result.get(HESS) is None:
        result[HESS] = sres_to_fim(result.get(SRES))

    return result


def reduce_result_via_options(
    result: ResultDict, options: HistoryOptions
) -> ResultDict:
    """Set values not to be stored in history or missing to NaN.

    Parameters
    ----------
    result:
        Result dictionary with all fields present.
    options:
        History options.

    Returns
    -------
    result:
        Result reduced to what is intended to be stored in history.
    """
    result = result.copy()

    # apply options to result
    for key in HistoryBase.RESULT_KEYS:
        if result.get(key) is None or not options.get(
            f'trace_record_{key}', True
        ):
            result[key] = np.nan

    return result



import logging
from typing import Iterable

import numpy as np
from scipy.stats import multivariate_normal

from ..problem import Problem
from ..result import ProfilerResult, Result
from .util import initialize_profile

logger = logging.getLogger(__name__)


def approximate_parameter_profile(
    problem: Problem,
    result: Result,
    profile_index: Iterable[int] = None,
    profile_list: int = None,
    result_index: int = 0,
    n_steps: int = 100,
) -> Result:
    """
    Calculate profile approximation.

    Based on an approximation via a normal likelihood centered at the chosen
    optimal parameter value, with the covariance matrix being the Hessian or
    FIM.

    Parameters
    ----------
    problem:
        The problem to be solved.
    result:
        A result object to initialize profiling and to append the profiling
        results to. For example, one might append more profiling runs to a
        previous profile, in order to merge these.
        The existence of an optimization result is obligatory.
    profile_index:
        List with the profile indices to be computed
        (by default all of the free parameters).
    profile_list:
        Integer which specifies whether a call to the profiler should create
        a new list of profiles (default) or should be added to a specific
        profile list.
    result_index:
        Index from which optimization result profiling should be started
        (default: global optimum, i.e., index = 0).
    n_steps:
        Number of profile steps in each dimension.

    Returns
    -------
    result:
        The profile results are filled into `result.profile_result`.
    """
    # Handling defaults
    # profiling indices
    if profile_index is None:
        profile_index = problem.x_free_indices

    # create the profile result object (retrieve global optimum) or append to
    # existing list of profiles
    global_opt = initialize_profile(
        problem, result, result_index, profile_index, profile_list
    )

    # extract optimization result
    optimizer_result = result.optimize_result.list[result_index]
    # extract values of interest
    x = optimizer_result.x
    fval = optimizer_result.fval
    hess = problem.get_reduced_matrix(optimizer_result.hess)

    # ratio scaling factor
    ratio_scaling = np.exp(global_opt - fval)

    # we need the hessian - compute if not provided or fishy
    if hess is None or np.isnan(hess).any():
        logger.info("Computing Hessian/FIM as not available in result.")
        hess = problem.objective(
            problem.get_reduced_vector(x), sensi_orders=(2,)
        )

    # inverse of the hessian
    sigma = np.linalg.inv(hess)

    # the steps
    xs = np.linspace(problem.lb_full, problem.ub_full, n_steps).T

    # loop over parameters for profiling
    for i_par in profile_index:
        # not requested or fixed -> compute no profile
        if i_par in problem.x_fixed_indices:
            continue

        i_free_par = problem.full_index_to_free_index(i_par)

        ys = multivariate_normal.pdf(
            xs[i_par], mean=x[i_par], cov=sigma[i_free_par, i_free_par]
        )

        fvals = -np.log(ys)
        ratios = ys / ys.max() * ratio_scaling

        profiler_result = ProfilerResult(
            x_path=xs,
            fval_path=fvals,
            ratio_path=ratios,
        )

        result.profile_result.set_profiler_result(
            profiler_result=profiler_result,
            i_par=i_par,
            profile_list=profile_list,
        )

    return result



"""Auto-saving."""

import binascii
import datetime
import os
from typing import Callable, Union

from ..result import Result
from .save_to_hdf5 import write_result


def autosave(
    filename: Union[str, Callable, None],
    result: Result,
    store_type: str,
    overwrite: bool = False,
):
    """
    Save the result of optimization, profiling or sampling automatically.

    Parameters
    ----------
    filename:
        Either the filename to save to or "Auto", in which case it will
        automatically generate a file named
        `year_month_day_{type}_result.hdf5`.
        A method can also be provided. All input to the autosave method will
        be passed to the filename method. The output should be the filename
        (`str`).
    result:
        The result to be saved.
    store_type:
        Either `optimize`, `sample` or `profile`. Depending on the
        method the function is called in.
    overwrite:
        Whether to overwrite the currently existing results.
    """
    if filename is None:
        return

    if filename == "Auto":
        filename = default_filename
    if not isinstance(filename, str):
        filename = filename(
            result=result,
            store_type=store_type,
            overwrite=overwrite,
        )
    # set the type to True and pass it on to write_result
    to_save = {store_type: True}
    write_result(
        result=result, overwrite=overwrite, filename=filename, **to_save
    )


def default_filename(**kwargs) -> str:
    """Create a filename when results will be autosaved.

    See :func:`autosave` for additional information.
    """
    time = datetime.datetime.now().strftime("%Y_%d_%m_%H_%M_%S")
    filename = (
        time + f"_{kwargs['store_type']}_result_"
        f"{binascii.b2a_hex(os.urandom(8)).decode()}.h5"
    )
    return filename



import logging
from typing import Callable, Iterable, Union

from ..engine import Engine, SingleCoreEngine
from ..optimize import Optimizer
from ..problem import Problem
from ..result import Result
from ..store import autosave
from .options import ProfileOptions
from .profile_next_guess import next_guess
from .task import ProfilerTask
from .util import initialize_profile

logger = logging.getLogger(__name__)


def parameter_profile(
    problem: Problem,
    result: Result,
    optimizer: Optimizer,
    engine: Engine = None,
    profile_index: Iterable[int] = None,
    profile_list: int = None,
    result_index: int = 0,
    next_guess_method: Union[Callable, str] = 'adaptive_step_regression',
    profile_options: ProfileOptions = None,
    progress_bar: bool = True,
    filename: Union[str, Callable, None] = None,
    overwrite: bool = False,
) -> Result:
    """
    Call to do parameter profiling.

    Parameters
    ----------
    problem:
        The problem to be solved.
    result:
        A result object to initialize profiling and to append the profiling
        results to. For example, one might append more profiling runs to a
        previous profile, in order to merge these.
        The existence of an optimization result is obligatory.
    optimizer:
        The optimizer to be used along each profile.
    engine:
        The engine to be used.
    profile_index:
        List with the parameter indices to be profiled
        (by default all free indices).
    profile_list:
        Integer which specifies whether a call to the profiler should create
        a new list of profiles (default) or should be added to a specific
        profile list.
    result_index:
        Index from which optimization result profiling should be started
        (default: global optimum, i.e., index = 0).
    next_guess_method:
        Function handle to a method that creates the next starting point for
        optimization in profiling.
    profile_options:
        Various options applied to the profile optimization.
    progress_bar:
        Whether to display a progress bar.
    filename:
        Name of the hdf5 file, where the result will be saved. Default is
        None, which deactivates automatic saving. If set to
        "Auto" it will automatically generate a file named
        `year_month_day_profiling_result.hdf5`.
        Optionally a method, see docs for `pypesto.store.auto.autosave`.
    overwrite:
        Whether to overwrite `result/profiling` in the autosave file
        if it already exists.

    Returns
    -------
    result:
        The profile results are filled into `result.profile_result`.
    """
    # Handling defaults
    # profiling indices
    if profile_index is None:
        profile_index = problem.x_free_indices

    # check profiling options
    if profile_options is None:
        profile_options = ProfileOptions()
    profile_options = ProfileOptions.create_instance(profile_options)

    # create a function handle that will be called later to get the next point
    if isinstance(next_guess_method, str):

        def create_next_guess(
            x,
            par_index,
            par_direction_,
            profile_options_,
            current_profile_,
            problem_,
            global_opt_,
        ):
            return next_guess(
                x,
                par_index,
                par_direction_,
                profile_options_,
                next_guess_method,
                current_profile_,
                problem_,
                global_opt_,
            )

    elif callable(next_guess_method):
        raise Exception(
            'Passing function handles for computation of next '
            'profiling point is not yet supported.'
        )
    else:
        raise Exception('Unsupported input for next_guess_method.')

    # create the profile result object (retrieve global optimum) or append to
    # existing list of profiles
    global_opt = initialize_profile(
        problem, result, result_index, profile_index, profile_list
    )
    # if engine==None set SingleCoreEngine() as default
    if engine is None:
        engine = SingleCoreEngine()

    # create Tasks
    tasks = []
    # loop over parameters to create tasks
    for i_par in profile_index:
        # only compute profiles for free parameters
        if i_par in problem.x_fixed_indices:
            continue

        current_profile = result.profile_result.get_profiler_result(
            i_par=i_par,
            profile_list=profile_list,
        )

        task = ProfilerTask(
            current_profile=current_profile,
            problem=problem,
            optimizer=optimizer,
            options=profile_options,
            create_next_guess=create_next_guess,
            global_opt=global_opt,
            i_par=i_par,
        )
        tasks.append(task)

    # execute the tasks with Engine
    indexed_profiles = engine.execute(tasks, progress_bar=progress_bar)

    # fill in the ProfilerResults at the right index
    for indexed_profile in indexed_profiles:
        result.profile_result.list[-1][
            indexed_profile['index']
        ] = indexed_profile['profile']

    autosave(
        filename=filename,
        result=result,
        store_type="profile",
        overwrite=overwrite,
    )

    return result



"""Contains the PetabImporter class."""
from __future__ import annotations

import importlib
import logging
import os
import shutil
import sys
import tempfile
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Union,
)

import numpy as np
import pandas as pd

from ..C import CONDITION_SEP, MODE_FUN, MODE_RES
from ..hierarchical.calculator import HierarchicalAmiciCalculator
from ..hierarchical.problem import InnerProblem
from ..objective import AggregatedObjective, AmiciObjective
from ..objective.amici import AmiciObjectBuilder
from ..objective.priors import NegLogParameterPriors, get_parameter_prior_dict
from ..predict import AmiciPredictor
from ..problem import Problem
from ..result import PredictionResult
from ..startpoint import FunctionStartpoints, StartpointMethod

try:
    import amici
    import amici.parameter_mapping
    import amici.petab_import
    import amici.petab_objective
    import petab
    from petab.C import PREEQUILIBRATION_CONDITION_ID, SIMULATION_CONDITION_ID
except ImportError:
    pass

logger = logging.getLogger(__name__)


class PetabImporter(AmiciObjectBuilder):
    """
    Importer for Petab files.

    Create an `amici.Model`, an `objective.AmiciObjective` or a
    `pypesto.Problem` from Petab files.
    """

    MODEL_BASE_DIR = "amici_models"

    def __init__(
        self,
        petab_problem: petab.Problem,
        output_folder: str = None,
        model_name: str = None,
        validate_petab: bool = True,
        hierarchical: bool = False,
    ):
        """Initialize importer.

        Parameters
        ----------
        petab_problem:
            Managing access to the model and data.
        output_folder:
            Folder to contain the amici model. Defaults to
            './amici_models/{model_name}'.
        model_name:
            Name of the model, which will in particular be the name of the
            compiled model python module.
        validate_petab:
            Flag indicating if the PEtab problem shall be validated.
        hierarchical:
            Whether to use hierarchical optimization or not, in case the
            underlying PEtab problem has parameters marked for hierarchical
            optimization (non-empty `parameterType` column in the PEtab
            parameter table).
        """
        self.petab_problem = petab_problem
        self._hierarchical = hierarchical

        if validate_petab:
            if petab.lint_problem(petab_problem):
                raise ValueError("Invalid PEtab problem.")
            if self._hierarchical:
                from ..hierarchical.petab import (
                    validate_hierarchical_petab_problem,
                )

                validate_hierarchical_petab_problem(petab_problem)

        if output_folder is None:
            output_folder = _find_output_folder_name(
                self.petab_problem,
                model_name=model_name,
            )
        self.output_folder = output_folder

        if model_name is None:
            model_name = _find_model_name(self.output_folder)
        self.model_name = model_name

    @staticmethod
    def from_yaml(
        yaml_config: Union[dict, str],
        output_folder: str = None,
        model_name: str = None,
    ) -> PetabImporter:
        """Simplified constructor using a petab yaml file."""
        petab_problem = petab.Problem.from_yaml(yaml_config)

        return PetabImporter(
            petab_problem=petab_problem,
            output_folder=output_folder,
            model_name=model_name,
        )

    def check_gradients(
        self,
        *args,
        rtol: float = 1e-2,
        atol: float = 1e-3,
        mode: Union[str, List[str]] = None,
        multi_eps=None,
        **kwargs,
    ) -> bool:
        """
        Check if gradients match finite differences (FDs).

        Parameters
        ----------
        rtol: relative error tolerance
        atol: absolute error tolerance
        mode: function values or residuals
        objAbsoluteTolerance: absolute tolerance in sensitivity calculation
        objRelativeTolerance: relative tolerance in sensitivity calculation
        multi_eps: multiple test step width for FDs

        Returns
        -------
        match: Whether gradients match FDs (True) or not (False)
        """
        par = np.asarray(self.petab_problem.x_nominal_scaled)
        problem = self.create_problem()
        objective = problem.objective
        free_indices = par[problem.x_free_indices]
        dfs = []
        modes = []

        if mode is None:
            modes = [MODE_FUN, MODE_RES]
        else:
            modes = [mode]

        if multi_eps is None:
            multi_eps = np.array([10 ** (-i) for i in range(3, 9)])

        for mode in modes:
            try:
                dfs.append(
                    objective.check_grad_multi_eps(
                        free_indices,
                        *args,
                        **kwargs,
                        mode=mode,
                        multi_eps=multi_eps,
                    )
                )
            except (RuntimeError, ValueError):
                # Might happen in case PEtab problem not well defined or
                # fails for specified tolerances in forward sensitivities
                return False

        return all(
            [
                any(
                    [
                        np.all(
                            (mode_df.rel_err.values < rtol)
                            | (mode_df.abs_err.values < atol)
                        ),
                    ]
                )
                for mode_df in dfs
            ]
        )

    def create_model(
        self,
        force_compile: bool = False,
        **kwargs,
    ) -> amici.Model:
        """
        Import amici model.

        Parameters
        ----------
        force_compile:
            If False, the model is compiled only if the output folder does not
            exist yet. If True, the output folder is deleted and the model
            (re-)compiled in either case.

            .. warning::
                If `force_compile`, then an existing folder of that name will
                be deleted.

        kwargs: Extra arguments passed to amici.SbmlImporter.sbml2amici
        """
        # courtesy check whether target is folder
        if os.path.exists(self.output_folder) and not os.path.isdir(
            self.output_folder
        ):
            raise AssertionError(
                f"Refusing to remove {self.output_folder} for model "
                f"compilation: Not a folder."
            )

        # add module to path
        if self.output_folder not in sys.path:
            sys.path.insert(0, self.output_folder)

        # compile
        if self._must_compile(force_compile):
            logger.info(
                f"Compiling amici model to folder " f"{self.output_folder}."
            )
            self.compile_model(**kwargs)
        else:
            logger.debug(
                f"Using existing amici model in folder "
                f"{self.output_folder}."
            )

        return self._create_model()

    def _create_model(self) -> amici.Model:
        """Load model module and return the model, no checks/compilation."""
        # load moduĺe
        module = amici.import_model_module(
            module_name=self.model_name, module_path=self.output_folder
        )
        model = module.getModel()
        amici.petab_import.check_model(
            amici_model=model,
            petab_problem=self.petab_problem,
        )

        return model

    def _must_compile(self, force_compile: bool):
        """Check whether the model needs to be compiled first."""
        # asked by user
        if force_compile:
            return True

        # folder does not exist
        if not os.path.exists(self.output_folder) or not os.listdir(
            self.output_folder
        ):
            return True

        # try to import (in particular checks version)
        try:
            # importing will already raise an exception if version wrong
            importlib.import_module(self.model_name)
        except ModuleNotFoundError:
            return True
        except amici.AmiciVersionError as e:
            logger.info(
                "amici model will be re-imported due to version "
                f"mismatch: {e}"
            )
            return True

        # no need to (re-)compile
        return False

    def compile_model(self, **kwargs):
        """
        Compile the model.

        If the output folder exists already, it is first deleted.

        Parameters
        ----------
        kwargs: Extra arguments passed to `amici.SbmlImporter.sbml2amici`.

        """
        # delete output directory
        if os.path.exists(self.output_folder):
            shutil.rmtree(self.output_folder)

        amici.petab_import.import_model(
            petab_problem=self.petab_problem,
            model_name=self.model_name,
            model_output_dir=self.output_folder,
            **kwargs,
        )

    def create_solver(self, model: amici.Model = None) -> amici.Solver:
        """Return model solver."""
        # create model
        if model is None:
            model = self.create_model()

        solver = model.getSolver()
        return solver

    def create_edatas(
        self, model: amici.Model = None, simulation_conditions=None
    ) -> List[amici.ExpData]:
        """Create list of amici.ExpData objects."""
        # create model
        if model is None:
            model = self.create_model()

        return amici.petab_objective.create_edatas(
            amici_model=model,
            petab_problem=self.petab_problem,
            simulation_conditions=simulation_conditions,
        )

    def create_objective(
        self,
        model: amici.Model = None,
        solver: amici.Solver = None,
        edatas: Sequence[amici.ExpData] = None,
        force_compile: bool = False,
        **kwargs,
    ) -> AmiciObjective:
        """Create a :class:`pypesto.AmiciObjective`.

        Parameters
        ----------
        model:
            The AMICI model.
        solver:
            The AMICI solver.
        edatas:
            The experimental data in AMICI format.
        force_compile:
            Whether to force-compile the model if not passed.
        **kwargs:
            Additional arguments passed on to the objective.

        Returns
        -------
        objective:
            A :class:`pypesto.AmiciObjective` for the model and the data.
        """
        # get simulation conditions
        simulation_conditions = petab.get_simulation_conditions(
            self.petab_problem.measurement_df
        )

        # create model
        if model is None:
            model = self.create_model(force_compile=force_compile)
        # create solver
        if solver is None:
            solver = self.create_solver(model)
        # create conditions and edatas from measurement data
        if edatas is None:
            edatas = self.create_edatas(
                model=model, simulation_conditions=simulation_conditions
            )

        parameter_mapping = amici.petab_objective.create_parameter_mapping(
            petab_problem=self.petab_problem,
            simulation_conditions=simulation_conditions,
            scaled_parameters=True,
            amici_model=model,
            fill_fixed_parameters=False,
        )

        par_ids = self.petab_problem.x_ids

        # fill in dummy parameters (this is needed since some objective
        #  initialization e.g. checks for preeq parameters)
        problem_parameters = {
            key: val
            for key, val in zip(
                self.petab_problem.x_ids, self.petab_problem.x_nominal_scaled
            )
        }
        amici.parameter_mapping.fill_in_parameters(
            edatas=edatas,
            problem_parameters=problem_parameters,
            scaled_parameters=True,
            parameter_mapping=parameter_mapping,
            amici_model=model,
        )

        calculator = None
        amici_reporting = None
        if self._hierarchical:
            inner_problem = InnerProblem.from_petab_amici(
                self.petab_problem, model, edatas
            )
            calculator = HierarchicalAmiciCalculator(inner_problem)
            amici_reporting = amici.RDataReporting.full
            inner_parameter_ids = calculator.inner_problem.get_x_ids()
            par_ids = [x for x in par_ids if x not in inner_parameter_ids]

            # FIXME: currently not supported with hierarchical
            kwargs['guess_steadystate'] = False

        # create objective
        obj = AmiciObjective(
            amici_model=model,
            amici_solver=solver,
            edatas=edatas,
            x_ids=par_ids,
            x_names=par_ids,
            parameter_mapping=parameter_mapping,
            amici_object_builder=self,
            calculator=calculator,
            amici_reporting=amici_reporting,
            **kwargs,
        )

        return obj

    def create_predictor(
        self,
        objective: AmiciObjective = None,
        amici_output_fields: Sequence[str] = None,
        post_processor: Union[Callable, None] = None,
        post_processor_sensi: Union[Callable, None] = None,
        post_processor_time: Union[Callable, None] = None,
        max_chunk_size: Union[int, None] = None,
        output_ids: Sequence[str] = None,
        condition_ids: Sequence[str] = None,
    ) -> AmiciPredictor:
        """Create a :class:`pypesto.predict.AmiciPredictor`.

        The `AmiciPredictor` facilitates generation of predictions from
        parameter vectors.

        Parameters
        ----------
        objective:
            An objective object, which will be used to get model simulations
        amici_output_fields:
            keys that exist in the return data object from AMICI, which should
            be available for the post-processors
        post_processor:
            A callable function which applies postprocessing to the simulation
            results. Default are the observables of the AMICI model.
            This method takes a list of ndarrays (as returned in the field
            ['y'] of amici ReturnData objects) as input.
        post_processor_sensi:
            A callable function which applies postprocessing to the
            sensitivities of the simulation results. Default are the
            observable sensitivities of the AMICI model.
            This method takes two lists of ndarrays (as returned in the
            fields ['y'] and ['sy'] of amici ReturnData objects) as input.
        post_processor_time:
            A callable function which applies postprocessing to the timepoints
            of the simulations. Default are the timepoints of the amici model.
            This method takes a list of ndarrays (as returned in the field
            ['t'] of amici ReturnData objects) as input.
        max_chunk_size:
            In some cases, we don't want to compute all predictions at once
            when calling the prediction function, as this might not fit into
            the memory for large datasets and models.
            Here, the user can specify a maximum number of conditions, which
            should be simulated at a time.
            Default is 0 meaning that all conditions will be simulated.
            Other values are only applicable, if an output file is specified.
        output_ids:
            IDs of outputs, if post-processing is used
        condition_ids:
            IDs of conditions, if post-processing is used

        Returns
        -------
        predictor:
            A :class:`pypesto.predict.AmiciPredictor` for the model, using
            the outputs of the AMICI model and the timepoints from the
            PEtab data
        """
        # if the user didn't pass an objective function, we create it first
        if objective is None:
            objective = self.create_objective()

        # create a identifiers of preequilibration and simulation condition ids
        # which can then be stored in the prediction result
        edata_conditions = (
            objective.amici_object_builder.petab_problem.get_simulation_conditions_from_measurement_df()
        )
        if PREEQUILIBRATION_CONDITION_ID not in list(edata_conditions.columns):
            preeq_dummy = [''] * edata_conditions.shape[0]
            edata_conditions[PREEQUILIBRATION_CONDITION_ID] = preeq_dummy
        edata_conditions.drop_duplicates(inplace=True)

        if condition_ids is None:
            condition_ids = [
                edata_conditions.loc[id, PREEQUILIBRATION_CONDITION_ID]
                + CONDITION_SEP
                + edata_conditions.loc[id, SIMULATION_CONDITION_ID]
                for id in edata_conditions.index
            ]

        # wrap around AmiciPredictor
        predictor = AmiciPredictor(
            amici_objective=objective,
            amici_output_fields=amici_output_fields,
            post_processor=post_processor,
            post_processor_sensi=post_processor_sensi,
            post_processor_time=post_processor_time,
            max_chunk_size=max_chunk_size,
            output_ids=output_ids,
            condition_ids=condition_ids,
        )

        return predictor

    def create_prior(self) -> Union[NegLogParameterPriors, None]:
        """
        Create a prior from the parameter table.

        Returns None, if no priors are defined.
        """
        prior_list = []

        if petab.OBJECTIVE_PRIOR_TYPE in self.petab_problem.parameter_df:

            for i, x_id in enumerate(self.petab_problem.x_ids):

                prior_type_entry = self.petab_problem.parameter_df.loc[
                    x_id, petab.OBJECTIVE_PRIOR_TYPE
                ]

                if (
                    isinstance(prior_type_entry, str)
                    and prior_type_entry != petab.PARAMETER_SCALE_UNIFORM
                ):

                    prior_params = [
                        float(param)
                        for param in self.petab_problem.parameter_df.loc[
                            x_id, petab.OBJECTIVE_PRIOR_PARAMETERS
                        ].split(';')
                    ]

                    scale = self.petab_problem.parameter_df.loc[
                        x_id, petab.PARAMETER_SCALE
                    ]

                    prior_list.append(
                        get_parameter_prior_dict(
                            i, prior_type_entry, prior_params, scale
                        )
                    )

        if len(prior_list):
            return NegLogParameterPriors(prior_list)
        else:
            return None

    def create_startpoint_method(self) -> Union[StartpointMethod, None]:
        """Create a startpoint method.

        If the PEtab problem specifies an initializationPrior. Returns None,
        if no initializationPrior is specified.
        """
        if (
            petab.INITIALIZATION_PRIOR_TYPE
            not in self.petab_problem.parameter_df
        ):
            return None

        def startpoint_method(n_starts: int, **kwargs):
            return petab.sample_parameter_startpoints(
                self.petab_problem.parameter_df, n_starts=n_starts
            )

        return FunctionStartpoints(function=startpoint_method)

    def create_problem(
        self,
        objective: AmiciObjective = None,
        x_guesses: Optional[Iterable[float]] = None,
        problem_kwargs: Dict[str, Any] = None,
        **kwargs,
    ) -> Problem:
        """Create a :class:`pypesto.Problem`.

        Parameters
        ----------
        objective:
            Objective as created by `create_objective`.
        x_guesses:
            Guesses for the parameter values, shape (g, dim), where g denotes
            the number of guesses. These are used as start points in the
            optimization.
        problem_kwargs:
            Passed to the `pypesto.Problem` constructor.
        **kwargs:
            Additional key word arguments passed on to the objective,
            if not provided.

        Returns
        -------
        problem:
            A :class:`pypesto.Problem` for the objective.
        """
        if objective is None:
            objective = self.create_objective(**kwargs)

        x_fixed_indices = self.petab_problem.x_fixed_indices
        x_fixed_vals = self.petab_problem.x_nominal_fixed_scaled
        x_ids = self.petab_problem.x_ids
        lb = self.petab_problem.lb_scaled
        ub = self.petab_problem.ub_scaled

        # In case of hierarchical optimization, parameters estimated in the
        # inner subproblem are removed from the outer problem
        if self._hierarchical:
            if not isinstance(
                objective.calculator, HierarchicalAmiciCalculator
            ):
                raise AssertionError()
            inner_parameter_ids = (
                objective.calculator.inner_problem.get_x_ids()
            )
            lb = [b for x, b in zip(x_ids, lb) if x not in inner_parameter_ids]
            ub = [b for x, b in zip(x_ids, ub) if x not in inner_parameter_ids]
            x_ids = [x for x in x_ids if x not in inner_parameter_ids]
            x_fixed_indices = list(
                map(x_ids.index, self.petab_problem.x_fixed_ids)
            )

        x_scales = [
            self.petab_problem.parameter_df.loc[x_id, petab.PARAMETER_SCALE]
            for x_id in x_ids
        ]

        if problem_kwargs is None:
            problem_kwargs = {}

        prior = self.create_prior()

        if prior is not None:
            if self._hierarchical:
                raise NotImplementedError(
                    "Hierarchical optimization in combination "
                    "with priors is not yet supported."
                )
            objective = AggregatedObjective([objective, prior])

        problem = Problem(
            objective=objective,
            lb=lb,
            ub=ub,
            x_fixed_indices=x_fixed_indices,
            x_fixed_vals=x_fixed_vals,
            x_guesses=x_guesses,
            x_names=x_ids,
            x_scales=x_scales,
            x_priors_defs=prior,
            **problem_kwargs,
        )

        return problem

    def rdatas_to_measurement_df(
        self,
        rdatas: Sequence[amici.ReturnData],
        model: amici.Model = None,
    ) -> pd.DataFrame:
        """
        Create a measurement dataframe in the petab format.

        Parameters
        ----------
        rdatas:
            A list of rdatas as produced by
            pypesto.AmiciObjective.__call__(x, return_dict=True)['rdatas'].
        model:
            The amici model.

        Returns
        -------
        measurement_df:
            A dataframe built from the rdatas in the format as in
            self.petab_problem.measurement_df.
        """
        # create model
        if model is None:
            model = self.create_model()

        measurement_df = self.petab_problem.measurement_df

        return amici.petab_objective.rdatas_to_measurement_df(
            rdatas, model, measurement_df
        )

    def rdatas_to_simulation_df(
        self,
        rdatas: Sequence[amici.ReturnData],
        model: amici.Model = None,
    ) -> pd.DataFrame:
        """
        See `rdatas_to_measurement_df`.

        Execpt a petab simulation dataframe is created, i.e. the measurement
        column label is adjusted.
        """
        return self.rdatas_to_measurement_df(rdatas, model).rename(
            columns={petab.MEASUREMENT: petab.SIMULATION}
        )

    def prediction_to_petab_measurement_df(
        self,
        prediction: PredictionResult,
        predictor: AmiciPredictor = None,
    ) -> pd.DataFrame:
        """
        Cast prediction into a dataframe.

        If a PEtab problem is simulated without post-processing, then the
        result can be cast into a PEtab measurement or simulation dataframe

        Parameters
        ----------
        prediction:
            A prediction result as produced by an AmiciPredictor
        predictor:
            The AmiciPredictor function

        Returns
        -------
        measurement_df:
            A dataframe built from the rdatas in the format as in
            self.petab_problem.measurement_df.
        """
        # create rdata-like dicts from the prediction result
        @dataclass
        class FakeRData:
            ts: np.ndarray
            y: np.ndarray

        rdatas = [
            FakeRData(ts=condition.timepoints, y=condition.output)
            for condition in prediction.conditions
        ]

        # add an AMICI model, if possible
        model = None
        if predictor is not None:
            model = predictor.amici_objective.amici_model

        return self.rdatas_to_measurement_df(rdatas, model)

    def prediction_to_petab_simulation_df(
        self,
        prediction: PredictionResult,
        predictor: AmiciPredictor = None,
    ) -> pd.DataFrame:
        """
        See `prediction_to_petab_measurement_df`.

        Except a PEtab simulation dataframe is created, i.e. the measurement
        column label is adjusted.
        """
        return self.prediction_to_petab_measurement_df(
            prediction, predictor
        ).rename(columns={petab.MEASUREMENT: petab.SIMULATION})


def _find_output_folder_name(
    petab_problem: petab.Problem,
    model_name: str,
) -> str:
    """
    Find a name for storing the compiled amici model in.

    If available, use the sbml model name from the `petab_problem` or the
    provided `model_name` (latter is given priority), otherwise create a
    unique name. The folder will be located in the
    `PetabImporter.MODEL_BASE_DIR` subdirectory of the current directory.
    """
    # check whether location for amici model is a file
    if os.path.exists(PetabImporter.MODEL_BASE_DIR) and not os.path.isdir(
        PetabImporter.MODEL_BASE_DIR
    ):
        raise AssertionError(
            f"{PetabImporter.MODEL_BASE_DIR} exists and is not a directory, "
            f"thus cannot create a directory for the compiled amici model."
        )

    # create base directory if non-existent
    if not os.path.exists(PetabImporter.MODEL_BASE_DIR):
        os.makedirs(PetabImporter.MODEL_BASE_DIR)

    # try sbml model id
    sbml_model_id = petab_problem.sbml_model.getId()
    if model_name is not None:
        sbml_model_id = model_name

    if sbml_model_id:
        output_folder = os.path.abspath(
            os.path.join(PetabImporter.MODEL_BASE_DIR, sbml_model_id)
        )
    else:
        # create random folder name
        output_folder = os.path.abspath(
            tempfile.mkdtemp(dir=PetabImporter.MODEL_BASE_DIR)
        )
    return output_folder


def _find_model_name(output_folder: str) -> str:
    """Just re-use the last part of the output folder."""
    return os.path.split(os.path.normpath(output_folder))[-1]



from typing import List, Optional, Union

import matplotlib.cm as cm
import numpy as np

from pypesto.util import assign_clusters

# for typehints
from ..C import RGBA


def assign_clustered_colors(vals, balance_alpha=True, highlight_global=True):
    """
    Cluster and assign colors.

    Parameters
    ----------
    vals: numeric list or array
        List to be clustered and assigned colors.
    balance_alpha: bool (optional)
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting (default: True)
    highlight_global: bool (optional)
        flag indicating whether global optimum should be highlighted

    Returns
    -------
    colors: list of RGBA
        One for each element in 'vals'.
    """
    # sanity checks
    if vals is None or len(vals) == 0:
        return []

    # assign clusters
    clusters, cluster_size = assign_clusters(vals)

    # create list of colors, which has the correct shape
    n_clusters = 1 + max(clusters) - sum(cluster_size == 1)

    # if best value was found more than once: we need one color less
    if highlight_global and cluster_size[0] > 1:
        n_clusters -= 1

    # fill color array from colormap
    colormap = cm.ScalarMappable().to_rgba
    color_list = colormap(np.linspace(0.0, 1.0, n_clusters))

    # best optimum should be colored in red
    if highlight_global and cluster_size[0] > 1:
        color_list = np.concatenate(([[1.0, 0.0, 0.0, 1.0]], color_list))

    # We have clustered the results. However, clusters may have size 1,
    # so we need to rearrange the regroup the results into "no_clusters",
    # which will be grey, and "real_clusters", which will be colored

    # run through the cluster_indices and collect which element
    # is in 1-element-clusters, and which is in real clusters
    no_clusters = np.where(cluster_size == 1)[0]
    real_clusters = np.unique(np.where(cluster_size > 1)[0])

    # assign transparency valuesaccording to cluster size, if wanted
    if balance_alpha:
        # assign neutral color, add 1 for avoiding division by zero
        grey = [0.7, 0.7, 0.7, min(1.0, 5.0 / (no_clusters.size + 1.0))]

        # reduce alpha level depend on size of each cluster
        n_cluster_size = np.delete(cluster_size, no_clusters)
        for icluster in range(n_clusters):
            color_list[icluster][3] = min(1.0, 5.0 / n_cluster_size[icluster])
    else:
        # assign neutral color
        grey = [0.7, 0.7, 0.7, 1.0]

    # create a color list, prfilled with grey values
    colors = np.array([grey] * clusters.size)

    # assign colors to real clusters
    for icol, iclust in enumerate(real_clusters):
        # find indices belonging to the cluster iclust and assign color
        ind_of_iclust = np.argwhere(clusters == iclust).flatten()
        colors[ind_of_iclust, :] = color_list[icol, :]

    # if best value was found only once: replace it with red
    if highlight_global and cluster_size[0] == 1:
        colors[0] = [1.0, 0.0, 0.0, 1.0]

    return colors


def assign_colors(
    vals, colors=None, balance_alpha=True, highlight_global=True
):
    """
    Assign colors or format user specified colors.

    Parameters
    ----------
    vals: numeric list or array
        List to be clustered and assigned colors.

    colors: list, or RGBA, optional
        list of colors, or single color

    balance_alpha: bool (optional)
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting (default: True)

    highlight_global: bool (optional)
        flag indicating whether global optimum should be highlighted

    Returns
    -------
    colors: list of RGBA
        One for each element in 'vals'.
    """
    # sanity checks
    if vals is None or len(vals) == 0:
        return np.array([])

    # if the user did not specify any colors:
    if colors is None:
        return assign_clustered_colors(
            vals,
            balance_alpha=balance_alpha,
            highlight_global=highlight_global,
        )

    # The user passed values and colors: parse them first!
    # we want everything to be numpy arrays, to not check every time whether a
    # list was passed or an ndarray
    colors = np.array(colors)

    # Get number of elements and use user assigned colors
    n_vals = len(vals) if isinstance(vals, list) else vals.size

    # Two usages are possible: One color for the whole data set, or one
    # color for each value:
    if colors.size == 4:
        # Only one color was passed: flatten in case and repeat n_vals times
        if colors.ndim == 2:
            colors = colors[0]
        return np.array([colors] * n_vals)

    if colors.shape[1] == 4 and n_vals == colors.shape[0]:
        return colors

    if colors.shape[0] == 4:
        colors = np.transpose(colors)
        if n_vals == colors.shape[0]:
            return colors

    # Shape of array did not match n_vals. Error due to size mismatch:
    raise ValueError(
        'Incorrect color input. Colors must be specified either as '
        'list of `[r, g, b, alpha]` with length equal to that of `vals` '
        f'(here: {n_vals}), or as a single `[r, g, b, alpha]`.'
    )


def assign_colors_for_list(
    num_entries: int,
    colors: Optional[Union[RGBA, List[RGBA], np.ndarray]] = None,
) -> Union[List[List[float]], np.ndarray]:
    """
    Create a list of colors for a list of items.

    Can also check a user-provided list of colors and use this if
    everything is ok.

    Parameters
    ----------
    num_entries:
        number of results in list
    colors:
        list of colors, or single color

    Returns
    -------
    colors:
        List of RGBA, one for each element in 'vals'.
    """
    # if the user did not specify any colors:
    if colors is None:
        # default colors will be used, on for each entry in the result list.
        # Colors are created from assign_colors, which needs a dummy list
        dummy_clusters = np.array(list(range(num_entries)) * 2)

        # we don't want alpha levels for all plotting routines in this case...
        colors = assign_colors(
            dummy_clusters, balance_alpha=False, highlight_global=False
        )

        # dummy cluster had twice as many entries as really there. Reduce.
        real_indices = np.arange(int(colors.shape[0] / 2))
        return colors[real_indices]

    # if the user specified color lies does not match the number of results
    if len(colors) != num_entries:
        raise (
            'Incorrect color input. Colors must be specified either as '
            'list of [r, g, b, alpha] with length equal to function '
            'values Number of function (here: ' + str(num_entries) + '), '
            'or as one single [r, g, b, alpha] color.'
        )

    return colors



"""CSV history."""

import copy
import os
import time
from typing import Dict, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from ..C import (
    FVAL,
    GRAD,
    HESS,
    N_FVAL,
    N_GRAD,
    N_HESS,
    N_RES,
    N_SRES,
    RES,
    SRES,
    TIME,
    ModeType,
    X,
)
from .base import CountHistoryBase, add_fun_from_res, reduce_result_via_options
from .options import HistoryOptions
from .util import MaybeArray, ResultDict, trace_wrap


class CsvHistory(CountHistoryBase):
    """Stores a representation of the history in a CSV file.

    Parameters
    ----------
    file:
        CSV file name.
    x_names:
        Parameter names.
    options:
        History options.
    load_from_file:
        If True, history will be initialized from data in the specified file
    """

    def __init__(
        self,
        file: str,
        x_names: Sequence[str] = None,
        options: Union[HistoryOptions, Dict] = None,
        load_from_file: bool = False,
    ):
        super().__init__(options=options)
        self.x_names: Sequence[str] = x_names
        self._trace: Union[pd.DataFrame, None] = None
        self.file: str = os.path.abspath(file)

        # create trace file dirs
        if self.file is not None:
            dirname = os.path.dirname(self.file)
            os.makedirs(dirname, exist_ok=True)

        if load_from_file and os.path.exists(self.file):
            trace = pd.read_csv(self.file, header=[0, 1], index_col=0)
            # replace 'nan' in cols with np.NAN
            cols = pd.DataFrame(trace.columns.to_list())
            cols[cols == 'nan'] = np.NaN
            trace.columns = pd.MultiIndex.from_tuples(
                cols.to_records(index=False).tolist()
            )
            for col in trace.columns:
                # transform strings to np.ndarrays
                trace[col] = trace[col].apply(string2ndarray)

            self._trace = trace
            self.x_names = trace[X].columns
            self._update_counts_from_trace()

    def _update_counts_from_trace(self) -> None:
        self._n_fval = self._trace[(N_FVAL, np.NaN)].max()
        self._n_grad = self._trace[(N_GRAD, np.NaN)].max()
        self._n_hess = self._trace[(N_HESS, np.NaN)].max()
        self._n_res = self._trace[(N_RES, np.NaN)].max()
        self._n_sres = self._trace[(N_SRES, np.NaN)].max()

    def update(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        """See `History` docstring."""
        super().update(x, sensi_orders, mode, result)
        self._update_trace(x, mode, result)

    def finalize(self, message: str = None, exitflag: str = None):
        """See `HistoryBase` docstring."""
        super().finalize()
        self._save_trace(finalize=True)

    def _update_trace(
        self,
        x: np.ndarray,
        mode: ModeType,
        result: ResultDict,
    ):
        """Update and possibly store the trace."""
        if not self.options.trace_record:
            return

        # init trace
        if self._trace is None:
            self._init_trace(x)

        # calculating function values from residuals
        #  and reduce via requested history options
        result = reduce_result_via_options(
            add_fun_from_res(result), self.options
        )

        used_time = time.time() - self._start_time

        # create table row
        row = pd.Series(
            name=len(self._trace), index=self._trace.columns, dtype='object'
        )

        values = {
            TIME: used_time,
            N_FVAL: self._n_fval,
            N_GRAD: self._n_grad,
            N_HESS: self._n_hess,
            N_RES: self._n_res,
            N_SRES: self._n_sres,
            FVAL: result[FVAL],
            RES: result[RES],
            HESS: result[HESS],
        }

        for var, val in values.items():
            row[(var, np.nan)] = val

        for var, val in {
            X: x,
            GRAD: result[GRAD],
        }.items():
            if var == X or self.options[f'trace_record_{var}']:
                row[var] = val
            else:
                row[(var, np.nan)] = np.nan

        self._trace = pd.concat(
            (self._trace, pd.DataFrame([row])),
        )

        # save trace to file
        self._save_trace()

    def _init_trace(self, x: np.ndarray):
        """Initialize the trace."""
        if self.x_names is None:
            self.x_names = [f'x{i}' for i, _ in enumerate(x)]

        columns: List[Tuple] = [
            (c, np.nan)
            for c in [
                TIME,
                N_FVAL,
                N_GRAD,
                N_HESS,
                N_RES,
                N_SRES,
                FVAL,
                RES,
                SRES,
                HESS,
            ]
        ]

        for var in [X, GRAD]:
            if var == X or self.options[f'trace_record_{var}']:
                columns.extend([(var, x_name) for x_name in self.x_names])
            else:
                columns.extend([(var,)])

        # TODO: multi-index for res, sres, hess
        self._trace = pd.DataFrame(
            columns=pd.MultiIndex.from_tuples(columns), dtype='float64'
        )

        # only non-float64
        trace_dtypes = {
            RES: 'object',
            SRES: 'object',
            HESS: 'object',
            N_FVAL: 'int64',
            N_GRAD: 'int64',
            N_HESS: 'int64',
            N_RES: 'int64',
            N_SRES: 'int64',
        }

        for var, dtype in trace_dtypes.items():
            self._trace[(var, np.nan)] = self._trace[(var, np.nan)].astype(
                dtype
            )

    def _save_trace(self, finalize: bool = False):
        """
        Save to file via pd.DataFrame.to_csv().

        Only done, if `self.storage_file` is not None and other conditions.
        apply.
        """
        if self.file is None:
            return

        if finalize or (
            len(self._trace) > 0
            and len(self._trace) % self.options.trace_save_iter == 0
        ):
            # save
            trace_copy = copy.deepcopy(self._trace)
            for field in [(HESS, np.nan), (RES, np.nan), (SRES, np.nan)]:
                trace_copy[field] = trace_copy[field].apply(
                    ndarray2string_full
                )
            trace_copy.to_csv(self.file)

    def __len__(self) -> int:
        """Define length of history object."""
        return len(self._trace)

    @trace_wrap
    def get_x_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        """See `HistoryBase` docstring."""
        return list(self._trace[X].values[ix])

    @trace_wrap
    def get_fval_trace(
        self, ix: Union[int, Sequence[int], None], trim: bool = False
    ) -> Union[Sequence[float], float]:
        """See `HistoryBase` docstring."""
        return list(self._trace[(FVAL, np.nan)].values[ix])

    @trace_wrap
    def get_grad_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return list(self._trace[GRAD].values[ix])

    @trace_wrap
    def get_hess_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return list(self._trace[(HESS, np.nan)].values[ix])

    @trace_wrap
    def get_res_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return list(self._trace[(RES, np.nan)].values[ix])

    @trace_wrap
    def get_sres_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return list(self._trace[(SRES, np.nan)].values[ix])

    @trace_wrap
    def get_time_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:
        """See `HistoryBase` docstring."""
        return list(self._trace[(TIME, np.nan)].values[ix])


def ndarray2string_full(x: Union[np.ndarray, None]) -> Union[str, None]:
    """
    Convert numpy array to string.

    Use 16-digit numerical precision and no truncation for large arrays.

    Parameters
    ----------
    x: array to convert.

    Returns
    -------
    x: array as string.
    """
    if not isinstance(x, np.ndarray):
        return x
    return np.array2string(
        x, threshold=x.size, precision=16, max_line_width=np.inf
    )


def string2ndarray(x: Union[str, float]) -> Union[np.ndarray, float]:
    """
    Convert string to numpy array.

    Parameters
    ----------
    x: array to convert.

    Returns
    -------
    x: array as np.ndarray.
    """
    if not isinstance(x, str):
        return x
    if x.startswith('[['):
        return np.vstack(
            [np.fromstring(xx, sep=' ') for xx in x[2:-2].split(']\n [')]
        )
    else:
        return np.fromstring(x[1:-1], sep=' ')
